import math

import numpy as np


def unit_vector(vector):
    return vector / np.linalg.norm(vector)


def angle_between(v1, v2):
    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)
    return np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))


def angle_between_points(p0, p1, p2):
    v0 = np.array(p0) - np.array(p1)
    v1 = np.array(p2) - np.array(p1)

    angle = np.math.atan2(np.linalg.det([v0, v1]), np.dot(v0, v1))
    return np.degrees(angle)


def angle_relative_to_180(angle):
    if angle < 0:
        return -180 - angle
    else:
        return 180 - angle


def angle(pt1,pt2):
    m1 = (pt1.getY() - pt1.getY())/1
    m2 = (pt2.getY() - pt1.getY())/(pt2.getX()-pt1.getX())

    tnAngle = (m1-m2)/(1+(m1*m2))
    return math.atan(tnAngle)


import numpy as np
import cv2


def gaussian_2d():
    """
    Create a 2-dimensional isotropic Gaussian map.
    :return: a 2D Gaussian map. 1000x1000.
    """
    mean = 0
    radius = 2.5
    # a = 1 / (2 * np.pi * (radius ** 2))
    a = 1.
    x0, x1 = np.meshgrid(np.arange(-5, 5, 0.01), np.arange(-5, 5, 0.01))
    x = np.append([x0.reshape(-1)], [x1.reshape(-1)], axis=0).T

    m0 = (x[:, 0] - mean) ** 2
    m1 = (x[:, 1] - mean) ** 2
    gaussian_map = a * np.exp(-0.5 * (m0 + m1) / (radius ** 2))
    gaussian_map = gaussian_map.reshape(len(x0), len(x1))

    max_prob = np.max(gaussian_map)
    min_prob = np.min(gaussian_map)
    gaussian_map = (gaussian_map - min_prob) / (max_prob - min_prob)
    gaussian_map = np.clip(gaussian_map, 0., 1.)
    return gaussian_map


class GaussianGenerator:
    def __init__(self):
        self.gaussian_img = gaussian_2d()

    @staticmethod
    def perspective_transform(src, dst_shape, dst_points):
        """
        Perspective Transform
        :param src: Image to transform.
        :param dst_shape: Tuple of 2 intergers(rows and columns).
        :param dst_points: [[x1, y1], [x2, y2], [x3, y3], [x4, y4]].
        :return: Image after perspective transform.
        """
        img = src.copy()
        h, w = img.shape[:2]
        src_points = np.float32([[0, 0], [w, 0], [w, h], [0, h]])
        dst_points = np.float32(dst_points)
        perspective_mat = cv2.getPerspectiveTransform(src=src_points, dst=dst_points)
        perspective_mat = np.array(perspective_mat)
        dst = cv2.warpPerspective(img, perspective_mat, (dst_shape[1], dst_shape[0]),
                                  borderValue=0, borderMode=cv2.BORDER_CONSTANT)
        dst = np.array(dst)
        return dst

    def gen(self, score_shape, points_list):
        score_map = np.zeros(score_shape, dtype=np.float32)
        for points in points_list:
            tmp_score_map = self.perspective_transform(self.gaussian_img, score_shape, points)
            score_map = np.where(tmp_score_map > score_map, tmp_score_map, score_map)
        score_map = np.clip(score_map, 0, 1.)
        # score_map = score_map.astype(np.uint8)
        return score_map


import base64
import io
import math
import pickle
import re
import glob
import json
import inspect
import sys

from copy import deepcopy
import numpy as np
import random
import os
import cv2
from imageio import imread
from tqdm import tqdm

IR_FRAMEWORK_MODE_HEADLESS = os.getenv("IR_FRAMEWORK_MODE_HEADLESS", None) is not None
if not IR_FRAMEWORK_MODE_HEADLESS:
    import tkinter.font as tkFont
    import tkinter as tk


def unify_string(string: str) -> str:
    """
    Converts the input string to lowercase and removes all special characters.

    Args:
        string: string to convert

    Returns: String in lowercase with all special characters removed.
    """
    lower = string.lower()
    return re.sub(r"[^a-z0-9]", "", lower)


def unify_path(path: str) -> str:
    """
    OpenCV may have problems opening images on a mixed slashes path.
    This function replaces forward slashes with double backwards slashes on windows systems
    and backwards on a unix system.
    Args:
        path: path string

    Returns: Path string in the format of the current operating system.

    """
    if os.name == "nt":
        return re.sub(r"/", "\\\\", path)
    else:
        return re.sub(r"\\", "/", path)


def isnan(arr):
    """
    Checks if the input is nan or infinite.

    Args:
        arr: Input to check

    Returns:
        True if the input is nan or infinite, else False.
    """
    if isinstance(arr, list):
        return np.any([isnan(x) for x in arr])
    return np.isnan(np.sum(arr)) or np.isinf(np.sum(arr))


def to_num(arr):
    """
    Replaces nan with zero and infinite with a high finite number.

    Args:
        arr: Iterable list type or single attribute.

    Returns:
        Input with all nan and inf values replaced.
    """
    if isinstance(arr, list):
        return [to_num(x) for x in arr]
    return np.nan_to_num(arr, copy=False, nan=0.0, posinf=1, neginf=0.0)


def load_json_from_file(path: str):
    """
    Loads and parses a json file as a dictionary.

    Args:
        path: location of json file

    Returns: dictionary or none if json error or file does not exist.
    """
    try:
        return json.load(open(path))
    except Exception as e:
        print("Load json: " + str(path) + " failed: " + str(e))
        return None


def get_next_numeric_folder_in_dir(path: str, alt_folder_name: str):
    """
    Gets the path to a folder that can be created by appending an increasing number to alt_folder_name.

    Args:
        path: Path where the folder should be created.
        alt_folder_name: Name of the folder.

    Returns:
        Path of a folder that can be created, with the given name and a number appended to it.
    """
    sub_dir = path + alt_folder_name + str(len(glob.glob(path + alt_folder_name + '*'))) + '\\'
    i = 0
    while True:
        if os.path.exists(sub_dir):
            sub_dir = path + alt_folder_name + str(i) + '\\'
            i += 1
        else:
            return sub_dir


def weighted_choice(gens: list):
    """
    Weighted random choice based on propability.
    Args:
        gens: list of elements to be chosen of

    Returns: element of given list based on their probability

    """
    x = random.random()
    for i, elmt in enumerate(gens):
        if x <= gens[i].get_probability():
            return elmt
        x -= gens[i].get_probability()


def rand_minmax(val_min=0, val_max=1, size=1):
    """
    uniform random value between val_min and val_max

    Args:
        val_min: min random value
        val_max: max random value
        size: number of random values

    Returns: random floating point value
        (list of random floating point values if size > 1)

    """
    x = val_min + np.random.rand(size) * (val_max - val_min)
    if size == 1:
        x = x[0]
    return x


def blocks(files, size=65536):
    """
    Calculates the number of lines of a large text file.
    (source: https://stackoverflow.com/questions/9629179/python-counting-lines-in-a-huge-10gb-file-as-fast-as-possible)
    Args:
        files: source file
        size: block size to be read

    Returns: number of lines

    """
    while True:
        b = files.read(size)
        if not b: break
        yield b


def print_warning(txt: str):
    """
    Prints a warning in the color yellow to the console.

    Args:
        txt: String to print.

    Returns:
        None
    """
    print("\033[93m WARNING: " + txt + '\033[0m')


def convert_image_to_255(image):
    """
    Converts the image from 0-1 to 0-255 if the image is in the 0-1 format, otherwise returns the image as it is.

    Args:
        image: image as array

    Returns:
        image as array with a number range from 0-255
    """
    if np.max(image) > 1 and not np.max(image) > 255:
        return image
    elif not np.max(image) > 1:
        return image * 255
    else:
        print_warning(f'Utils.convert_image_to_255 received image with max value {np.max(image)}. '
                      f'Normalizing image to 255.')
        return normalize_image(image, 255)


def convert_images_to_255(images):
    """
    Converts a list of images to the 0-255 format.

    Args:
        images: List of images.

    Returns:
        List of images convert to 255.
    """
    for i in range(len(images)):
        images[i] = convert_image_to_255(images[i])
    return images


def normalize_image(image, value):
    """
    Normalizes the image to the given value.

    Args:
        image: image as array
        value: the value to normalize to (usually 255 or 1)

    Returns:
        Image as array normalized to the given value.
    """
    return image / (value / np.max(image))


def get_funcs_in_modules(module, submodule_depth=0) -> list:
    """
    Returns the functions declared as from xxx import xxx in the module. Can explore submodules via depth param
    Args:
        module: the module to look in for functions
        submodule_depth: how many submodules to explore

    Returns: A list of functions in the submodules
    """
    funcs = inspect.getmembers(module, inspect.isfunction)
    for i in range(submodule_depth):
        submods = inspect.getmembers(module, inspect.ismodule)
        for submod in submods:
            subfuncs = get_funcs_in_modules(submod, submodule_depth - 1)
            funcs.extend(subfuncs)
    return funcs


def keys_in_dict(dictionary: dict, keys: list):
    """
    Checks if the given list of keys is contained in the dictionary.

    Args:
        dictionary: Dict to check.
        keys: Keys to check for.

    Returns:
        True if all keys are contained in the dict, else false.
    """
    return all([key in dictionary for key in keys])


def keys_in_dicts(dicts: list, keys: list):
    """
    Checks if all keys are contained in all dicts.

    Args:
        dicts: List of dicts to check.
        keys: List of keys to check for.

    Returns:
        True if all keys are contained in all dicts, else false.
    """
    return all([keys_in_dict(dictionary, keys) for dictionary in dicts])


def get_value_in_nested_dict(d: dict, keys: list, not_found_return=None):
    """
    Returns a value in a nested dict if value is in dict else None
    Args:
        d: dictionairy
        keys: a list of keys e.g. ['value1','value'] equals d['value1]['value2']
    Returns: value if in dict else None
    """
    for key in keys:
        try:
            d = d[key]
        except KeyError:
            return not_found_return
    return d


def nested_set(dic, keys, value, create_missing=True):
    d = dic
    for key in keys[:-1]:
        if key in d:
            d = d[key]
        elif create_missing:
            d = d.setdefault(key, {})
        else:
            return dic
    if keys[-1] in d or create_missing:
        d[keys[-1]] = value
    return dic


def do_import(class_path: str):
    """
    Imports a class from a string. Modules can either be separated by a '.' or a '/'.
    Example: class_path = data_processing/dp_trees/tree/Tree will import the tree class.

    Args:
        class_path: Path to import

    Returns:
        The imported class
    """
    class_path = class_path.replace('/', '.')
    module, import_class = class_path.rsplit('.', 1)
    return getattr(__import__(module, fromlist=[import_class]), import_class)


def convert_base_64_to_ndarray(base64img: str):
    '''
    Converts a Base64 encoded image (as returned by our DB) to a ndarray grayscale image.

    Args:
        base64img: Base64 encoded image
        color: Color of the output image

    Returns:
        ndarray grayscale image
    '''
    if base64img is None:
        print_warning("recieved None as image")
        return None

    encoded_img = base64img.split(',')[1]
    decoded_img = np.array(imread(io.BytesIO(base64.b64decode(encoded_img))))
    swapped = decoded_img[..., [2, 1, 0]].copy()
    return swapped


def cal_distance(p1, p2):
    return math.sqrt(math.pow((p2[0] - p1[0]), 2) + math.pow((p2[1] - p1[1]), 2))


def get_all_line_coords(point1, point2, resolution=1):
    """
    Args:
        point1: Point to start the line from
        point2: Point at which the line should end
        resolution: Resolution of the points. Higher resolution results in better accuracy, lower resolution will make
        computation less expensive.

    Returns:
        Points representing a line between the given points.
    """
    dist = int(cal_distance(point1, point2) * resolution)
    if dist == 0:
        return []
    x_diff = (point1[0] - point2[0]) / dist
    y_diff = (point1[1] - point2[1]) / dist
    line_points = [[point1[0] - x_diff * i, point1[1] - y_diff * i] for i in range(dist)]
    return line_points


def get_min_distance(point, points):
    min_dist = [None, float('inf')]
    for p in points:
        dist = cal_distance(point, p)
        if dist < min_dist[1]:
            min_dist = [p, dist]
    return min_dist


def get_conversion_matrix_boxes(src_points, target_points):
    assert len(src_points) == len(target_points), "Target and source point count need to match."
    assert len(src_points) % 2 == 0, "There must be an even number of source points."
    src_boxes = []
    target_boxes = []
    for i in range(1, int(len(src_points) / 2)):
        points = [src_points[i * 2 - 2], src_points[i * 2], src_points[i * 2 + 1], src_points[i * 2 - 1]]
        src_boxes.append(points)
        target_boxes.append(
            [target_points[i * 2 - 2], target_points[i * 2], target_points[i * 2 + 1], target_points[i * 2 - 1]])
    return np.float32(src_boxes), np.float32(target_boxes)


def center_to_zero(box):
    return np.array([[0, 0], [box[1][0] - box[0][0], 0], [box[2][0] - box[3][0], box[2][1] - box[1][1]],
                     [0, box[3][1] - box[0][1]]], dtype="float32")


def warp_polygon_box(img, src_points):
    src_boxes, target_boxes = calc_poly_warp(src_points)
    dst = warp_with_boxes(img, src_boxes, target_boxes)
    return dst, target_boxes, src_boxes


def calc_poly_warp(src_points):
    point_pairs = [[src_points[0], src_points[-1]]]
    forward_index = 1
    reverse_index = -2
    while len(src_points) + 1 > forward_index - reverse_index:
        bottom_line = get_all_line_coords(src_points[reverse_index], src_points[reverse_index + 1])
        top_line = get_all_line_coords(src_points[forward_index], src_points[forward_index - 1])
        closest_top_line_point, min_top_line_distance = get_min_distance(src_points[reverse_index], top_line)
        closest_bottom_line_point, min_bottom_line_distance = get_min_distance(src_points[forward_index], bottom_line)
        if min_top_line_distance < min_bottom_line_distance:
            point_pairs.append([src_points[reverse_index], closest_top_line_point])
            if all(np.array(closest_top_line_point, dtype="int32") == src_points[forward_index]):
                forward_index += 1
            reverse_index -= 1
        else:
            point_pairs.append([src_points[forward_index], closest_bottom_line_point])
            if all(np.array(closest_bottom_line_point, dtype="int32") == src_points[reverse_index]):
                reverse_index -= 1
            forward_index += 1
    average_height = int(sum([cal_distance(pp[0], pp[1]) for pp in point_pairs]) / len(point_pairs))
    rel_x = 0
    rel_y = 0
    target_points = [[rel_x, rel_y], [rel_x, rel_y + average_height]]
    source_points = [point_pairs[0][0], point_pairs[0][1]]
    distance_to_0 = 0
    for i in range(1, len(point_pairs)):
        x_coord = rel_x + distance_to_0 + (int((cal_distance(point_pairs[i][0], point_pairs[i - 1][0]) +
                                                cal_distance(point_pairs[i][1], point_pairs[i - 1][1])) / 2))
        target_points.append([x_coord, rel_y])
        target_points.append([x_coord, rel_y + average_height])

        # Make sure the boxes are in the correct order
        if point_pairs[i][0][1] < point_pairs[i][1][1]:
            source_points.append(np.int32(point_pairs[i][0]))
            source_points.append(np.int32(point_pairs[i][1]))
        else:
            source_points.append(np.int32(point_pairs[i][1]))
            source_points.append(np.int32(point_pairs[i][0]))
        distance_to_0 += int(cal_distance(point_pairs[i][0], point_pairs[i - 1][0]))
    src_boxes, target_boxes = get_conversion_matrix_boxes(source_points, target_points)
    return src_boxes, target_boxes


def warp_with_boxes(img, src_boxes, target_boxes):
    dst = None
    for src_box, target_box in zip(src_boxes, target_boxes):
        if any([cal_distance(src_box[i1], src_box[i2]) <= 0 for i1, i2 in [(0, 1), (2, 3), (1, 2), (0, 3)]]):
            continue
        target_box = center_to_zero(target_box)
        width = round((cal_distance(target_box[0], target_box[1]) + cal_distance(target_box[2], target_box[3])) / 2)
        height = round((cal_distance(target_box[1], target_box[2]) + cal_distance(target_box[3], target_box[0])) / 2)
        perspective_mat = cv2.getPerspectiveTransform(src=src_box, dst=target_box)
        perspective_mat = np.array(perspective_mat)
        if dst is None:
            dst = cv2.warpPerspective(img, perspective_mat, (width, height))
            dst = np.array(dst)
        else:
            warped = cv2.warpPerspective(img, perspective_mat, (width, height))
            warped = np.array(warped)
            dst = np.append(dst, warped, axis=1)
    return dst


def crop_image(src, points, dst_height=None):
    """
    Crop heat map with points.
    :param src: 8-bit single-channel image (map).
    :param points: [[x1, y1], [x2, y2], [x3, y3], [x4, y4]].
    :return: dst_heat_map: Cropped image. 8-bit single-channel image (map) of heat map.
             src_points: [[x1, y1], [x2, y2], [x3, y3], [x4, y4]].
             dst_points: [[x1, y1], [x2, y2], [x3, y3], [x4, y4]].
    """
    src_image = src.copy()
    src_points = np.float32(points)
    width = round((cal_distance(points[0], points[1]) + cal_distance(points[2], points[3])) / 2)
    height = round((cal_distance(points[1], points[2]) + cal_distance(points[3], points[0])) / 2)
    if dst_height is not None:
        if width == 0:
            width += 1
        if height == 0:
            height += 1
        ratio = dst_height / min(width, height)
        width = int(width * ratio)
        height = int(height * ratio)
    crop_points = np.float32([[0, 0], [width, 0], [width, height], [0, height]])
    perspective_mat = cv2.getPerspectiveTransform(src=src_points, dst=crop_points)
    perspective_mat = np.array(perspective_mat)
    dst_heat_map = cv2.warpPerspective(src_image, perspective_mat, (width, height),
                                       borderValue=0, borderMode=cv2.BORDER_CONSTANT)
    dst_heat_map = np.array(dst_heat_map)
    return dst_heat_map, src_points, crop_points


def pad_to_min_shape(img: np.array, min_shape):
    ## (4) pad image to have same size
    padding = np.array([0, 0, 0, 0])
    # pad image to have minimum shape of min_shape
    pad_pixels = np.array(min_shape) - np.array(img.shape)[:-1]
    pad_pixels[pad_pixels < 0] = 0
    if np.any(pad_pixels > 0):
        top, bottom = pad_pixels[0] // 2, pad_pixels[0] - (pad_pixels[0] // 2)
        left, right = pad_pixels[1] // 2, pad_pixels[1] - (pad_pixels[1] // 2)
        padding = np.array([left, top, bottom, right])
        img = cv2.copyMakeBorder(
            img,
            top=top,
            bottom=bottom,
            left=left,
            right=right,
            borderType=cv2.BORDER_CONSTANT,
            value=[0, 0, 0]
        )
        img = np.array(img)
    return img, padding


def crop_polygon(img, points):
    pts = np.float32(points)
    ## (1) Crop the bounding rect
    rect = cv2.boundingRect(pts)
    rect = np.array(rect)
    x, y, w, h = rect
    cropped = img[y:y + h, x:x + w].copy()

    ## (2) make mask
    pts = pts - pts.min(axis=0)
    mask = np.zeros(cropped.shape[:2], np.uint8)
    cv2.drawContours(mask, [pts.astype("int32")], -1, (1, 1, 1), -1, cv2.LINE_AA)
    mask = np.array(mask)

    ## (3) do bit-op
    dst = cv2.bitwise_and(cropped, cropped, mask=mask)
    dst = np.array(dst)

    return dst


def tree_search(search_list, set_value_list, check_func, target_value, always_return=False):
    """
    Greedy function that is called recursively to create and execute a basic depth first search tree.

    Args:
        search_list: The list of possible solutions for this tree node.
        set_value_list: The list of values set by previous nodes, for final execution and check.
        check_func: Function used for checking the result. Gets set_value_list as arg.
        target_value: Expected result of the cehck_func.
        always_return: Will return the set_value_list extended by the search list if no solution was found.
    Returns:
        First found correct solution or none, if no solution was found.
    """
    if len(search_list) == 1:
        local_set_value_list = deepcopy(set_value_list)
        local_set_value_list.append(search_list[0])
        if check_func(local_set_value_list) == target_value:
            return local_set_value_list
    else:
        for option in search_list:
            local_search_list = deepcopy(search_list)
            local_set_value_list = deepcopy(set_value_list)
            local_set_value_list.append(option)
            local_search_list.remove(option)
            res = tree_search(local_search_list, local_set_value_list, check_func, target_value)
            if res is not None:
                return res
    if always_return:
        set_value_list.extend(search_list)
        return set_value_list


gettrace = getattr(sys, 'gettrace', None)


def check_debug():
    return os.getenv("project_x_debug") == "true"


def save_as_pickle(data, path):
    with open(path, 'wb') as file:
        pickle.dump(data, file)
    return data


def filter_kwargs(func, kwargs):
    filtered_call_params = {}
    for k, v in kwargs.items():
        if k in inspect.signature(func).parameters or any(str(p.kind) == 'VAR_KEYWORD' for p in
                                                          inspect.signature(func).parameters.values()):
            filtered_call_params[k] = v
    return filtered_call_params


def img_normalize_craft(src):
    """
    Normalize a RGB image.
    :param src: Image to normalize. Must be RGB order.
    :return: Normalized Image
    """
    NORMALIZE_MEAN = np.array([0.485, 0.456, 0.406], dtype=np.float32) * 255.0
    NORMALIZE_VARIANCE = np.array([0.229, 0.224, 0.225], dtype=np.float32) * 255.0
    img = src.copy().astype(np.float32)

    img -= NORMALIZE_MEAN
    img /= NORMALIZE_VARIANCE
    return img


class LabelPrintedLenCalc:
    def __init__(self, family="Arial", font_size=20, char_num=63488, location=None):
        self.known_chars = {}
        self.family = family
        self.font_size = font_size
        self.char_num = char_num
        if location is None:
            self.location = os.path.join(os.getcwd(), 'cache', 'utf8_char_map.json')
        else:
            self.location = location
        self.__get_char_lens_tkinter()

    def __get_char_lens_tkinter(self):
        if not os.path.exists(self.location):
            if IR_FRAMEWORK_MODE_HEADLESS:
                print("Framework is running in headless mode. Using fallback for char map.")
                from .__utf_char_map__ import get_char_map_fallback
                self.known_chars = get_char_map_fallback()
                return
            print("Generating utf char map. This may take a while.")
            self.generate_utf8_char_map()
        else:
            print(f"Using cached char map @ {self.location}")
        with open(self.location, 'r') as file:
            self.known_chars = json.load(file)
    
    def generate_utf8_char_map(self):
        root = tk.Tk()
        font = tkFont.Font(family=self.family, size=self.font_size)
        print(f'Calculating UTF8 char len map for n = {self.char_num} characters.')
        for i in tqdm(range(self.char_num)):
            try:
                char = chr(i)
                (w, h) = (font.measure(char), font.metrics("linespace"))
                self.known_chars[char] = w
            except UnicodeEncodeError or ValueError:
                continue
        os.makedirs(os.path.dirname(self.location), exist_ok=True)
        with open(self.location, 'w') as file:
            json.dump(self.known_chars, file, indent=4)

    def __calc_char_len(self, char):
        if char not in self.known_chars:
            return int(sum(self.known_chars.values()) / len(self.known_chars))
        return self.known_chars[char]

    def calc_label_len(self, label):
        return sum([self.__calc_char_len(char) for char in label])


def get_printed_text_size(text, predictor=LabelPrintedLenCalc()):
    return predictor.calc_label_len(text)


import cv2
import numpy as np
from numpy import random

from . import standard
from . import generator
from .bounding_box import rescale_bbox

def color_fix(image: np.array, target_shape):
    """
    Converts the the image into the color format given by the target shape.

    Args:
        image: Image to convert.
        target_shape: Target shape for the image.

    Returns:
        Image converted into the target color format.
    """
    # RGB to Gray
    if target_shape[-1] == 1 and len(image.shape) == 3 and image.shape[-1] == 3:
        image = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)
    # RGBA to Gray
    elif target_shape[-1] == 1 and len(image.shape) == 3 and image.shape[-1] == 4:
        image = cv2.cvtColor(image, cv2.COLOR_RGBA2GRAY)
    # GRAY to RGB
    elif target_shape[-1] == 3 and (len(image.shape) != 3 or image.shape[-1] == 1):
        image = cv2.cvtColor(image, cv2.COLOR_GRAY2RGB)
    # gray no channel to gray with channel
    elif target_shape[-1] == 1 and len(image.shape) == 2:
        image = image.reshape(target_shape)
    # RGBA to RGB
    elif target_shape[-1] == 3 and (len(image.shape) == 3 and image.shape[-1] == 4):
        image = cv2.cvtColor(image, cv2.COLOR_RGBA2RGB)

    image = np.array(image)
    return image


def preprocess_image(image: np.array, target_shape) -> np.array:
    """
    This method rescales the image to target images,
    makes sure that all batches have the same color bbox_format (of target shape)
    Replaces nan values
    And normalizes the image to 0 to 1 by divding all values by 255

    Args:
        image: the image to process
        target_shape: the output shape and bbox_format

    Returns:
        The preprocessed image as np.array
    """
    image = np.array(image)
    # rescale image
    image_size = target_shape[:-1]
    image_size_array = np.array(image_size)
    origin_image_size = np.array(image.shape)
    scale = image_size_array / origin_image_size[:2]
    image = cv2.resize(image, dsize=tuple(image_size[::-1]), interpolation=cv2.INTER_CUBIC)
    image = np.array(image)
    # fixing errors
    image = standard.to_num(image)
    # to flaot
    image = image.astype(np.float32)
    # divide by 255 to normalize image
    #NORMALIZE_MEAN = np.array([0.485, 0.456, 0.406], dtype=np.float32) * 255.0
    #NORMALIZE_VARIANCE = np.array([0.229, 0.224, 0.225], dtype=np.float32) * 255.0
    #image = image.copy().astype(np.float32)
    image /= 255
    #image -= NORMALIZE_MEAN
    #image /= NORMALIZE_VARIANCE
    # image bbox_format
    image = color_fix(image, target_shape)

    return image, scale


def get_and_preprocess_template(image, template_shape, bbox):
    """
    Gets and preprocesses a template, for example for target tracking.

    Args:
        image: the image to extract the template from
        template_shape: the target shape to which the template is resized to
        bbox: the bounding box of the template that gets extracted

    Returns:
        A preprocessed template cut out of the image
    """
    output_hw = template_shape[:-1][::-1]
    template = generator.get_transformation(image, output_hw[:2], bbox['bounding_box'].tolist())
    template = preprocess_image(template, template_shape[1:])[0]
    return template


def preprocess_images(images: list, target_shape, return_scale=True) -> np.array:
    """
    wrapper for preprocess_image for multiple images
    Args:
        images: a list of images as np.arrays
        target_shape: the shape the images shall have
        return_scale: Whether to return the reshaping scale (True) or the actual reshaped image (False).
    Returns:
        A list of preprocessed images or the scales the images need to be resized by if return_scale == True.
    """
    images_scales = np.array([preprocess_image(img, target_shape) for img in images], dtype=object)
    if return_scale:
        return images_scales
    return np.array([im_sc[0] for im_sc in images_scales])


def rescale_bounding_boxes(target_shape, image_shapes, bounding_boxes):
    for i in range(len(image_shapes)):
        image_shapes[i] = image_shapes[i] if len(image_shapes[i]) == 2 else image_shapes[i][:2]
    scales = target_shape[:2] / np.array(image_shapes)
    rescaled_bboxes = rescale_bounding_boxes_by_scales(bounding_boxes, scales)
    return rescaled_bboxes


def rescale_bounding_boxes_by_scales(boxes, scales):
    """

    Args:
        boxes: Boxes to rescale.
        scales: Scales as single tuple (x, y) or list of tuples [(x,y)..]
            If a single tuple is given all boxes are rescaled by the given factors.
            If a list of tuples is given, each box gets individually scaled by the given factor.

    Returns:
        Rescaled bounding boxes.
    """
    # No boxes. Return
    if not boxes or len(boxes) == 0 or (len(boxes[0]) == 0 and len(boxes) <= 1):
        return boxes

    # single scale which is applied to all
    scales = np.array(scales)
    if scales.shape == (2,):
        scales = np.array([scales for i in range(len(boxes))])

    # Note: Changed it from boxes[0][0] to boxes[0] if you have to change it again please contact Matthias.
    # Change was required because bounding box list is not nested anymore

    # Do it for four point bounding boxes
    if np.array(boxes[0]).shape == (4, 2):
        # tile scales to four point and then easy matrix multiplication
        scales = np.tile(scales, 4).reshape(len(scales), 4, 2)

    rescaled_boxes = [[np.array(box) * [scale[1], scale[0]] for box in img_boxes] for scale, img_boxes in zip(scales, boxes)]
    return rescaled_boxes

    # Two point bounding box
    #return [np.array(boxes[i]) * np.tile(scales[i], 2).T for i in range(len(boxes))]


def process_inputs(batch, target_shape):
    """
    Rescales the images in batch[0] and the corresponding bounding boxes.

    Args:
        batch: Batch returned by generator.
        target_shape: Target shape for the images.

    Returns:
        images: Rescaled images.
        rescaled_boxes: Rescaled bounding boxes.
    """
    image_scale = preprocess_images(batch[0], target_shape)
    images = np.array([isc[0] for isc in image_scale])
    # calculate scalses
    scales = np.array([isc[1] for isc in image_scale])
    rescaled_boxes = []
    for i, entry in enumerate(batch[1]):
        scale = [scales[i][1], scales[i][0]]
        boxes_for_entry = []
        for bbox in entry['bounding_box']:
            rescaled_bbox = rescale_bbox([bbox], scale)
            boxes_for_entry.append(rescaled_bbox)
        rescaled_boxes.append(boxes_for_entry)

    return images, rescaled_boxes


def preprocess_true_boxes(true_boxes, anchors, num_classes, input_shape, output_shapes):
    # Todo: Damit die Methode funktioniert muss herausgefunden werden, was die output shapes tun. Meines Wissens nach kommen diese aus den Outputs des Model, müssen also per routing aus der networkfactory gezogen werden.
    """
    Preprocess true boxes to training input bbox_format

    Args:
        true_boxes: Absolute x_min, y_min, x_max, y_max, class_id relative to input_shape.
        input_shape: (height, width), multiples of 32
        anchors: shape=(N, 2), (width, height)
        num_classes: Number of classes

    Returns:
        y_true: list of array, shape like yolo_outputs, xywh are relative values
    """
    assert (true_boxes[..., 4] < num_classes).all(), 'class id must be less than num_classes'
    num_layers = len(anchors) // 3  # default setting
    anchor_mask = [[6, 7, 8], [3, 4, 5], [0, 1, 2]] if num_layers == 3 else [[3, 4, 5], [0, 1, 2]]

    # Transform box info to (x_center, y_center, box_width, box_height, cls_id)
    # and image relative coordinate.
    true_boxes = np.array(true_boxes, dtype='float32')
    input_shape = np.array(input_shape[:-1], dtype='int32')

    boxes_xy = (true_boxes[..., 0:2] + true_boxes[..., 2:4]) // 2
    boxes_wh = true_boxes[..., 2:4] - true_boxes[..., 0:2]
    true_boxes[..., 0:2] = boxes_xy / input_shape[::-1]
    true_boxes[..., 2:4] = boxes_wh / input_shape[::-1]

    batch_size = true_boxes.shape[0]

    # grid_shapes = [input_shape // ]
    # TODO: Akzeptiert gerade nur die letzten num_layers als factor.
    #  Besser mehr unterstützen oder parametrisieren
    max_down_sample_factor = input_shape[0] // output_shapes[0][0]
    grid_shapes = [
        input_shape // {0: max_down_sample_factor, 1: max_down_sample_factor//2, 2: max_down_sample_factor//4}[l]
        for l in range(num_layers)
    ]

    yolo_boxes = [np.zeros((batch_size, grid_shapes[l][0], grid_shapes[l][1], len(anchor_mask[l]), 5 + num_classes),
                       dtype='float32') for l in range(num_layers)]

    # Expand dim to apply broadcasting.
    anchors = np.expand_dims(anchors, 0)
    anchor_maxes = anchors / 2.
    anchor_mins = -anchor_maxes
    valid_mask = boxes_wh[..., 0] > 0

    for b in range(batch_size):
        # Discard zero rows.
        wh = boxes_wh[b, valid_mask[b]]
        if len(wh) == 0:
            continue
        best_anchor = get_best_anchor(anchor_maxes, anchor_mins, anchors, wh)
        for t, n in enumerate(best_anchor):
            for l in range(num_layers):
                if n in anchor_mask[l]:
                    i = np.floor(true_boxes[b, t, 0] * grid_shapes[l][1]).astype('int32')
                    j = np.floor(true_boxes[b, t, 1] * grid_shapes[l][0]).astype('int32')
                    k = anchor_mask[l].index(n)
                    c = true_boxes[b, t, 4].astype('int32')
                    yolo_boxes[l][b, j, i, k, 0:4] = true_boxes[b, t, 0:4]
                    yolo_boxes[l][b, j, i, k, 4] = 1
                    yolo_boxes[l][b, j, i, k, 5 + c] = 1

    return yolo_boxes


def get_best_anchor(anchor_maxes, anchor_mins, anchors, wh):
    """
    Get the best yolo anchor for specific bounding boxes.

    Args:
        anchor_maxes: Max anchor values.
        anchor_mins: Min anchor values.
        anchors: Yolo anchors.
        wh: width and height of the boxes.

    Returns:
        The best anchor for the boxes.
    """
    # Expand dim to apply broadcasting.
    wh = np.expand_dims(wh, -2)

    box_maxes = wh / 2.
    box_mins = -box_maxes

    intersect_mins = np.maximum(box_mins, anchor_mins)
    intersect_maxes = np.minimum(box_maxes, anchor_maxes)
    intersect_wh = np.maximum(intersect_maxes - intersect_mins, 0.)
    intersect_area = intersect_wh[..., 0] * intersect_wh[..., 1]

    box_area = wh[..., 0] * wh[..., 1]
    anchor_area = anchors[..., 0] * anchors[..., 1]
    iou = intersect_area / (box_area + anchor_area - intersect_area)

    # Find best anchor for each true box
    best_anchor = np.argmax(iou, axis=-1)

    return best_anchor


def weighted_choice(items: list):
    """
    Weighted random choice based on probability. An item must have the attribute probability to be used

    Args:
        items: list of elements to be chosen of

    Returns:
        Element of given list based on their probability
    """
    if len(items) == 1:
        return items[0]

    x = random.random() * sum([item.probability for item in items])
    for i, element in enumerate(items):
        if x <= items[i].probability:
            return element
        x -= items[i].probability


def get_layer_shapes(layer_datapipe_mapping):
    """
    Gets the shapes of the layers.

    Args:
        layer_datapipe_mapping: List of neuron reference wrappings.

    Returns:
        Shapes of the neurons as func_dict with keys 0, 1, ..., n.
    """
    shapes = {}
    for layername in layer_datapipe_mapping:
        if layername.deref().shape not in shapes.items():
            shapes[len(shapes.keys())] = layername.deref().shape
    return shapes


def rescale_images(images, target_shape):
    """
    Rescales the images and returns the scales they were rescaled by.

    Args:
        images: Images to rescale.
        shapes: List of target images as func_dict with keys 0, 1, ..., n.

    Returns:
        images: Rescaled images as func_dict with the image shape as key.
        scales: The scales the images where rescaled by as func_dict with the image images as key.
    """
    images = [color_fix(img, (img.shape[0], img.shape[1], 3)) for img in images]
    images = preprocess_images(images, target_shape, return_scale=False)
    return images


def rescale_image(image, target_shape):
    """
    Rescales the images and returns the scales they were rescaled by.

    Args:
        images: Images to rescale.
        shapes: List of target images as func_dict with keys 0, 1, ..., n.

    Returns:
        images: Rescaled images as func_dict with the image shape as key.
        scales: The scales the images where rescaled by as func_dict with the image images as key.
    """
    image = color_fix(image, (image.shape[0], image.shape[1], 3))
    image, _ = preprocess_image(image, target_shape)
    return image


def filter_annotations_for_class(class_name):
    def _filter_annotations_for_class(batch):
        filtered = []
        for x in batch:
            filtered.append([ann for ann in x['annotations'] if ann['type'] == class_name])
        return filtered

    return _filter_annotations_for_class


def get_annotation_vals(query):
    def _f(annotations):
        anns = []
        for x in annotations:
            if type(query) == list:
                anns.append([
                    standard.get_value_in_nested_dict(ann, query) for ann in x
                ])
            else:
                anns.append([ann[query] for ann in x])
        return anns
    return _f


def get_reshaped_masks(target_shape, rescaled_bounding_boxes):
    """
    Creates masks and reshapes them.

    Args:
        batch: Batch for predictions.
        bbox_scales: Scales the boxes need to be rescaled by to match the original rescaled image.
        key: Neuron the masks should be created for.
        key_ref: Dict containing reference keys and the corresponding image images.
        kwargs: Additional arguments.
        images: Images the masks are created for.

    Returns:
        reshaped_masks: N Masks, N depending on the batch size.
    """
    masks = [generator.get_mask(target_shape[:2], rescaled_img_bounding_boxes)
             for rescaled_img_bounding_boxes in rescaled_bounding_boxes]
    return masks


def store_data_in_inputs_and_outputs(neurons, data, key_refs, inputs=None, outputs=None):
    """
    Stores the data in the input and/or output func_dict and returns them.

    Args:
        neurons: Neurons the function should be executed for.
        data: Data to store in inputs / outputs.
        key_refs: Keys referring to the shapes.
        inputs: Inputs passed to the network.
        outputs: Outputs of the network.

    Returns:
       inputs: Inputs passed to the network.
        outputs: Outputs of the network.
    """
    for neuron in neurons:
        for key in key_refs:
            shape = key_refs[key]
            if neuron.deref().shape == shape:
                if inputs is not None:
                    inputs[neuron] = data[key]
                if outputs is not None:
                    outputs[neuron] = data[key]
    return inputs, outputs


def get_template_dict(batch, shapes, targets):
    """
    Gets a dictionary containing the templates with a key ranging from 0...n.

    Args:
        batch: Batch for predictions.
        shapes: Shapes the templates should be created for.
        targets: Targets for the template generation (exactly one bounding box per image).

    Returns:
        Dict containing keys from 0...n and templates for each image and each shape.
    """
    results = {}
    for key in shapes:
        templates = []
        template_shape = shapes[key]
        for i in range(len(batch[0])):
            template = get_and_preprocess_template(batch[0][i], template_shape, targets[i])
            templates.append(np.array(template))
        results[key] = np.array(templates)
    return results


def convert_bb_to_yolov4bb(bounding_boxes, input_shapes, output_shapes, anchors=None):
    """
    Converts the standard bounding boxes to YoloV4 bounding boxes.

    Args:
        model: The keras model of the neural network
        neurons: The reference wrappings of the processing tree.
        bounding_boxes: The bounding boxes to convert to YoloV4 format.

    Returns:
        Bounding boxes in YoloV4 format.
    """
    # prepare to bbox_format where multiple boxes could be per image --> 0 is the class
    bbx = np.array([np.array([np.append(b, 0) for b in image]) for image in bounding_boxes])
    bbx = [image.reshape([1, image.shape[0], image.shape[1]]) for image in bbx]
    yolo_bounding_boxes = [preprocess_true_boxes(image, anchors, 1, input_shapes, output_shapes) for image in bbx]
    n1 = np.array([x[0][0] for x in yolo_bounding_boxes])
    n2 = np.array([x[1][0] for x in yolo_bounding_boxes])
    n3 = np.array([x[2][0] for x in yolo_bounding_boxes])
    # fixing nan erros in preprocessing
    yolo_bounding_boxes = [n1, n2, n3]
    yolo_bounding_boxes = standard.to_num(yolo_bounding_boxes)
    return yolo_bounding_boxes



import math
from copy import deepcopy

import cv2
import numpy as np

from . import bounding_box


def gaussian_2d():
    """
    Create a 2-dimensional isotropic Gaussian map.
    :return: a 2D Gaussian map. 1000x1000.
    """
    mean = 0
    radius = 1.4
    a = 1.
    x0, x1 = np.meshgrid(np.arange(-5, 5, 0.01), np.arange(-5, 5, 0.01))
    x = np.append([x0.reshape(-1)], [x1.reshape(-1)], axis=0).T

    m0 = (x[:, 0] - mean) ** 2
    m1 = (x[:, 1] - mean) ** 2
    gaussian_map = a * np.exp(-0.5 * (m0 + m1) / (radius ** 2))
    gaussian_map = gaussian_map.reshape(len(x0), len(x1))

    max_prob = np.max(gaussian_map)
    min_prob = np.min(gaussian_map)
    gaussian_map = (gaussian_map - min_prob) / (max_prob - min_prob)
    gaussian_map = np.clip(gaussian_map, 0., 1.)
    return gaussian_map


class GaussianGenerator:
    def __init__(self):
        self.gaussian_img = gaussian_2d()
        self.gaussian_maps = {1000: self.gaussian_img}

    @staticmethod
    def perspective_transform(src, dst_shape, dst_points):
        """
        Perspective Transform
        :param src: Image to transform.
        :param dst_shape: Tuple of 2 intergers(rows and columns).
        :param dst_points: [[x1, y1], [x2, y2], [x3, y3], [x4, y4]].
        :return: Image after perspective transform.
        """
        img = src.copy()
        h, w = img.shape[:2]
        src_points = np.float32([[0, 0], [w, 0], [w, h], [0, h]])
        dst_points = np.float32(dst_points)
        perspective_mat = cv2.getPerspectiveTransform(src=src_points, dst=dst_points)
        perspective_mat = np.array(perspective_mat)
        dst = cv2.warpPerspective(img, perspective_mat, (dst_shape[1], dst_shape[0]),
                                  borderValue=0, borderMode=cv2.BORDER_CONSTANT)
        dst = np.array(dst)
        return dst

    @staticmethod
    def normalize(img: np.ndarray):
        factor = 1 / img.max()
        return img * factor

    def get_gaus_map(self, shape):
        gaus_map_size = max(shape)
        if gaus_map_size not in self.gaussian_maps:
            gaus_map = self.normalize(cv2.resize(deepcopy(self.gaussian_img),
                                                                          (gaus_map_size, gaus_map_size)))
            self.gaussian_maps[gaus_map_size] = np.array(gaus_map)
        ret_value = cv2.resize(deepcopy(self.gaussian_maps[gaus_map_size]), shape, interpolation=cv2.INTER_LINEAR)
        ret_value = np.array(ret_value)
        return ret_value

    def gen(self, score_shape, points_list):
        score_map = np.zeros(score_shape, dtype=np.float32)
        # TODO: Use polygons
        boxes = [bounding_box.four_point_bb_to_2_point(points, round_values=True) for points in points_list]
        for box in boxes:
            point1 = [int(box[0]), int(box[1])]
            point2 = [int(box[2]), int(box[3])]
            for i in range(2):
                if point1[i] < 0:
                    point1[i] = 0
                if point2[i] < 0:
                    point2[i] = 0

            # assure not out of bounds and not identical start and end points
            point1[0] = np.clip(point1[0], 0, score_map.shape[0] - 1)
            point1[1] = np.clip(point1[1], 0, score_map.shape[1] - 1)
            point2[0] = np.clip(point2[0], 0, score_map.shape[0] - 1)
            point2[1] = np.clip(point2[1], 0, score_map.shape[1] - 1)

            x_points = np.array([point1[0], point2[0]])
            y_points = np.array([point1[1], point2[1]])
            if np.all(x_points >= score_map.shape[0]-2) or np.all(x_points <= 0) \
                    or np.all(y_points >= score_map.shape[1]-2) or np.all(y_points <= 0)\
                    or x_points[0] == x_points[1] or y_points[0] == y_points[1]:
                continue

            if min((point2[0] - point1[0], point2[1] - point1[1])) <= 3:
                if point1[0] - 1 > 0:
                    point1[0] -= 1
                if point1[1] - 1 > 0:
                    point1[1] -= 1
                if point2[0] + 1 < score_shape[0]:
                    point2[0] += 1
                if point2[1] + 1 < score_shape[1]:
                    point2[1] += 1

            # do gaussian transform
            img_section = score_map[point1[0]:point2[0], point1[1]:point2[1]]
            temp_gaus_map = self.get_gaus_map((img_section.shape[1], img_section.shape[0]))
            score_map[point1[0]:point2[0], point1[1]:point2[1]] = np.where(
                temp_gaus_map > img_section,
                temp_gaus_map,
                img_section
            )
        score_map = np.clip(score_map, 0, 1.)
        return score_map


def cal_quadrangle_area(points):
    """
    Calculate the area of quadrangle.
    :return: The area of quadrangle.
    """
    points = reorder_points(points)
    p1, p2, p3, p4 = points
    s1 = cal_triangle_area(p1, p2, p3)
    s2 = cal_triangle_area(p3, p4, p1)
    s3 = cal_triangle_area(p2, p3, p4)
    s4 = cal_triangle_area(p4, p1, p2)
    if s1 + s2 == s3 + s4:
        return s1 + s2
    else:
        return 0


def reorder_points(point_list):
    """
    Reorder points of quadrangle.
    (top-left, top-right, bottom right, bottom left).
    :param point_list: List of point. Point is (x, y).
    :return: Reorder points.
    """
    # Find the first point which x is minimum.
    if len(point_list) == 3:
        point_list = bounding_box.convert_triangle_to_quadrangle(point_list)
    elif len(point_list) == 2:
        point_list = bounding_box.two_point_bb_to_4_point(point_list)
    elif len(point_list) > 4:
        point_list = bounding_box.convert_polygon_box_to_4_point_bb(point_list)
    elif len(point_list) < 2:
        raise ValueError('Box passed to reorder points must have at least 2 points.')
    ordered_point_list = sorted(point_list, key=lambda x: (x[0], x[1]))
    first_point = ordered_point_list[0]

    # Find the third point. The slope is middle.
    slope_list = [[cal_slope(first_point, p), p] for p in ordered_point_list[1:]]
    ordered_slope_point_list = sorted(slope_list, key=lambda x: x[0])
    first_third_slope, third_point = ordered_slope_point_list[1]

    # Find the second point which is above the line between the first point and the third point.
    # All that's left is the fourth point.
    if above_line(ordered_slope_point_list[0][1], third_point, first_third_slope):
        second_point = ordered_slope_point_list[0][1]
        fourth_point = ordered_slope_point_list[2][1]
        reverse_flag = False
    else:
        second_point = ordered_slope_point_list[2][1]
        fourth_point = ordered_slope_point_list[0][1]
        reverse_flag = True

    # Find the top left point.
    second_fourth_slope = cal_slope(second_point, fourth_point)
    if first_third_slope < second_fourth_slope:
        if reverse_flag:
            reorder_point_list = [fourth_point, first_point, second_point, third_point]
        else:
            reorder_point_list = [second_point, third_point, fourth_point, first_point]
    else:
        reorder_point_list = [first_point, second_point, third_point, fourth_point]

    return reorder_point_list


def cal_triangle_area(p1, p2, p3):
    """
    Calculate the area of triangle.
    S = |(x2 - x1)(y3 - y1) - (x3 - x1)(y2 - y1)| / 2
    :param p1: (x, y)
    :param p2: (x, y)
    :param p3: (x, y)
    :return: The area of triangle.
    """
    [x1, y1], [x2, y2], [x3, y3] = p1, p2, p3
    return abs((x2 - x1) * (y3 - y1) - (x3 - x1) * (y2 - y1)) / 2


def cal_slope(p1, p2):
    return (p2[1] - p1[1]) / (p2[0] - p1[0] + 1e-5)


def above_line(p, start_point, slope):
    y = (p[0] - start_point[0]) * slope + start_point[1]
    return p[1] < y


def calc_affinities_from_batch(batch):
    return [get_img_aff_boxes(batch, img_num) for img_num in range(len(batch))]


def get_img_aff_boxes(batch, img_num):
    img_aff_boxes = []
    boxes = batch[img_num]["annotations"]
    hierarchy = batch[img_num]["hierarchy"]
    for word in hierarchy:
        char_boxes = get_boxes_by_id(boxes, word['to'])
        if len(char_boxes) < 2:
            continue
        aff_boxes = cal_affinity_boxes(char_boxes)
        img_aff_boxes.extend(aff_boxes)
    return img_aff_boxes


def get_boxes_by_id(boxes, ids):
    boxes_new = []
    for b_id in ids:
        for box in boxes:
            if box['id'] == b_id:
                boxes_new.append(bounding_box.convert_polygon_box_to_4_point_bb(box['box']))
                break
    return boxes_new


def cal_affinity_boxes(char_boxes):
    ordered_char_boxes = [reorder_points(cb) for cb in char_boxes]
    ordered_char_boxes = reorder_box(ordered_char_boxes)

    fix_box_dimensions(ordered_char_boxes)

    point_pairs_list = [cal_point_pairs(bbox) for bbox in ordered_char_boxes]
    aff_boxes = []
    for i in range(len(point_pairs_list) - 1):
        affinity_box = cal_affinity_box(point_pairs_list[i], point_pairs_list[i + 1])
        aff_boxes.append(affinity_box)

    return aff_boxes


def fix_box_dimensions(ordered_char_boxes):
    """
    Makes sure the box is at least 1 pixel in each direction to prevent bugs.

    Args:
        ordered_char_boxes: List of ordered char boxes to check.

    Returns:
        None, changes original list.
    """
    for box in ordered_char_boxes:
        if all([b[0] == box[0][0] for b in box]):
            box[1][0] += 1
            box[2][0] += 1
        if all([b[1] == box[0][1] for b in box]):
            box[2][1] += 1
            box[3][1] += 1


def reorder_box(box_list):
    """
    Reorder character boxes.
    :param box_list: List of box. Box is a list of point. Point is (x, y).
    :return: Reorder boxes.
    """
    # Calculate the minimum distance between any two boxes.
    box_count = len(box_list)
    distance_mat = np.zeros((box_count, box_count), dtype=np.float32)
    for i in range(box_count):
        box1 = box_list[i]
        for j in range(i + 1, box_count):
            box2 = box_list[j]
            distance = cal_min_box_distance(box1, box2)
            distance_mat[i][j] = distance
            distance_mat[j][i] = distance

    # Find the boxes on the both ends.
    end_box_index = np.argmax(distance_mat)
    nan = distance_mat[end_box_index // box_count, end_box_index % box_count] + 1
    for i in range(box_count):
        distance_mat[i, i] = nan
    last_box_index = start_box_index = end_box_index // box_count
    last_box = box_list[start_box_index]

    # reorder box.
    reordered_box_list = [last_box]
    for i in range(box_count - 1):
        distance_mat[:, last_box_index] = nan
        closest_box_index = np.argmin(distance_mat[last_box_index])
        reordered_box_list.append(box_list[closest_box_index])
        last_box_index = closest_box_index

    return reordered_box_list


def cal_point_pairs(points):
    intersection = intersection_of_diagonals(points)
    p1, p2, p3, p4 = points
    point_pairs = [[cal_center_point([p1, p2, intersection]), cal_center_point([p3, p4, intersection])],
                   [cal_center_point([p2, p3, intersection]), cal_center_point([p4, p1, intersection])]]
    return point_pairs


def cal_min_box_distance(box1, box2):
    box_distance = [math.pow((p2[0] - p1[0]), 2) + math.pow((p2[1] - p1[1]), 2) for p1 in box1 for p2 in box2]
    return np.min(box_distance)


def intersection_of_diagonals(points):
    """
    Calculate the intersection of diagonals.
    x=[(x3-x1)(x4-x2)(y2-y1)+x1(y3-y1)(x4-x2)-x2(y4-y2)(x3-x1)]/[(y3-y1)(x4-x2)-(y4-y2)(x3-x1)]
    y=(y3-y1)[(x4-x2)(y2-y1)+(x1-x2)(y4-y2)]/[(y3-y1)(x4-x2)-(y4-y2)(x3-x1)]+y1
    :param points: (x1, y1), (x2, y2), (x3, y3), (x4, y4).
    :return: (x, y).
    """
    [x1, y1], [x2, y2], [x3, y3], [x4, y4] = points
    x = ((x3 - x1) * (x4 - x2) * (y2 - y1) + x1 * (y3 - y1) * (x4 - x2) - x2 * (y4 - y2) * (x3 - x1)) \
        / ((y3 - y1) * (x4 - x2) - (y4 - y2) * (x3 - x1) + 1e-5)
    y = (y3 - y1) * ((x4 - x2) * (y2 - y1) + (x1 - x2) * (y4 - y2)) \
        / ((y3 - y1) * (x4 - x2) - (y4 - y2) * (x3 - x1) + 1e-5) + y1
    return [x, y]


def cal_center_point(points):
    points = np.array(points)
    return [round(np.average(points[:, 0])), round(np.average(points[:, 1]))]


def cal_affinity_box(point_pairs1, point_pairs2):
    areas = [cal_quadrangle_area([p1, p2, p3, p4]) for p1, p2 in point_pairs1 for p3, p4 in point_pairs2]
    max_area_index = np.argmax(areas)
    affinity_box = [point_pairs1[max_area_index // 2][0],
                    point_pairs1[max_area_index // 2][1],
                    point_pairs2[max_area_index % 2][0],
                    point_pairs2[max_area_index % 2][1]]
    return np.int32(affinity_box)


from PIL import Image, ImageChops
from . import bounding_box
import cv2
from .standard import unify_path


def get_mask(shape, bbox: list) -> np.array:
    """
    Returns a black image with the bounding box in white
    Args:
        shape: shape of the new image
        bbox: bounding box locations

    Returns:
        np.array all values are zero, bounding box values are 1
    """
    mask_img = np.zeros(shape, dtype="uint8")
    for bb in bbox:
        if len(bb) == 4:
            bb = bounding_box.order_points(bb)
        mask_img = cv2.fillPoly(mask_img, [np.array(bb).astype("int32")], color=255).astype("float32")
        mask_img = np.array(mask_img)
    return mask_img


def read_img_as_cv2(img_path: str) -> np.array:
    """
    Reads an image from disk into cv2 bbox_format
    Args:
        img_path: the path to the image on disk

    Returns:
        the loaded image as np.array
    """
    try:
        img = cv2.imread(unify_path(img_path), cv2.IMREAD_COLOR)
        if img is None:
            img = np.array(Image.open(img_path).convert("RGB"))
            img = cv2.cvtColor(img, cv2.IMREAD_COLOR)
    except Exception as e:
        print("Problems with  img: " + img_path + " Error: " + str(e))
        img = None
    return img


def resize_img(img: np.ndarray, dim=(512, 512)) -> np.ndarray:
    """
    Resizes an image by absolute dimensions or scale between 0.0 and 1.0.
    Args:
        img: image to resize
        dim: tuple - absolute scaling - (x,y)
             float - relative scaling - 0.0 < x <= 1.0

    Returns: resized image

    """
    if type(dim) is tuple:
        return cv2.resize(img, dim, interpolation=cv2.INTER_AREA)
    elif type(dim) is float:
        w = int(img.shape[1] * dim)
        h = int(img.shape[0] * dim)
        return cv2.resize(img, (w, h), interpolation=cv2.INTER_AREA)
    else:
        raise TypeError("Dimension type not supported. Provide tuple or float instead.")


def invert_img(img: np.ndarray) -> np.ndarray:
    """
    Inverts an image.
    Args:
        img: image to invert

    Returns: inverted image

    """
    return cv2.bitwise_not(img)


def gaussian_noise(img, mean=0, var=0.1):
    """
    Applies a gaussian noise effect to an image.
    Args:
        img: source
        mean: TODO: add mean description
        var: strength of noise

    Returns: image with noise

    """
    row, col = img.shape[:2]
    sigma = var ** 0.5
    gauss = np.random.normal(mean, sigma, (row, col, 3))
    gauss = gauss.reshape((row, col, 3))
    noisy = (img + gauss).astype(np.uint8)
    return noisy


def smoothen(img):
    """
    Smoothes an image.
    Args:
        img: source

    Returns: smoothed image

    """
    return cv2.GaussianBlur(img, (5, 5), 0)


def auto_crop_content(img):
    """
    Crops an image.
    Args:
        img: image (numpy array)

    Returns: cropped image (numpy array)

    """
    pil = True
    if type(img) is np.ndarray:
        pil = False
        im = Image.fromarray(img)
    else:
        im = img.copy()
    bg = Image.new(im.mode, im.size, im.getpixel((0, 0)))
    diff = ImageChops.difference(im, bg)
    diff = ImageChops.add(diff, diff, 2.0, -100)
    bbox = diff.getbbox()
    if bbox:
        bbox = (bbox[0] - 2, bbox[1] - 2, bbox[2] + 2, bbox[3] + 2)
        return im.crop(bbox)
    if pil:
        return im
    else:
        return np.array(im)


def get_wh(bbox):
    """
    Calculated width and height from bounding box.
    Args:
        bbox: two point bounding box

    Returns: width and height

    """
    if len(bbox) > 2:
        # 4-point or poly
        return bbox[1][0] - bbox[0][0], bbox[2][1] - bbox[0][1]
    # 2-point
    return bbox[1][0] - bbox[0][0], bbox[1][1] - bbox[0][1]


def get_transformation(img: np.ndarray, output_shape, bbox: list) -> np.ndarray:
    """
    Based on the bounding box that image part is cut out and then resized to output_shape
    Args:
        img: Image as numpy array
        output_shape: The shape the cutout template gets resized to (height, width)
        bbox: the bounding box for the part of the image

    Returns:
        the resized cutout
    """
    raw_image = bounding_box.cut_out_image_area(img, bbox[0])
    return cv2.resize(raw_image, tuple(output_shape), interpolation=cv2.INTER_CUBIC)







from copy import deepcopy

from grandpa.template_parser import TemplateParser
from grandpa.node import Node
from grandpa.routing import Router
from grandpa.utils.standard import print_warning


class GraphRuntime(Node):
    """
    Core of the framework. Manages all graph functionalities.
    """
    def __init__(self, **kwargs):
        self.router = Router()
        super().__init__(address="runtime", main_router=self.router)
        self.graphs = {}

    def add_graph(self, graph_name: str, template: dict, template_settings: dict):
        """
        Adds a new graph to the runtime.

        Args:
            graph_name: Name of the graph.
            template: Template that should be used to create the graph.
            template_settings: Settings that should be used for the graph.

        Returns:
            Executable graph.
        """
        if graph_name in self.graphs:
            assert "GraphName {graph_name} already defined. Did you mean merge_graph?"

        new_graph = TemplateParser(template, template_settings, self.router)
        new_graph.initialized_nodes.append(self.address)
        self.graphs[graph_name] = new_graph

        return new_graph

    def get_graph_of_node(self, node):
        """
        Args:
            node: Node to get the graph of.

        Returns:
            Executable graph the given node is part of.
        """
        for graph in self.graphs.values():
            for graph_node in graph.initialized_nodes:
                graph_node = self.router.get_instruction('//' + graph_node)
                if node is graph_node:
                    return graph
        print_warning(f'{node.address} was not found in any graph of this runtime.')

    def update_graph(self, graph_to_update: str, update_with: str):
        """

        Args:
            graph_to_update: Graph to update.
            update_with: Graph to update the graph with.

        Returns:
            A graph that represents the new graph_to_update, a merge f both graphs.
        """
        graph_1 = self.graphs[graph_to_update]
        graph_2 = self.graphs[update_with]
        new_nodes, removed_nodes, nodes_with_changes = self.__get_differences(graph_1, graph_2)
        for node in new_nodes:
            graph_1.init_node(node.name, node.orig_node_def)
        for node in removed_nodes:
            self.router.delete_object_at_path('//' + node.name)
        for node in nodes_with_changes:
            if node.update_settings or node.update_setting_nodes:
                self.__reinitialize_node(node, graph_1)
            elif node.update_params or node.update_param_nodes:
                self.__update_node_params(node, graph_1)
        return graph_1

    def __update_node_params(self, node, graph):
        target_node = self.router.get_instruction('//' + node.name)
        target_node.nodes = {}
        graph.add_nodes(target_node, node.orig_node_def['params'], node.address)

    def __reinitialize_node(self, node, graph):
        self.router.delete_object_at_path('//' + node.name)
        graph.init_node(node.name, node.orig_node_def)

    def merge_graph(self, graph_name_1: str, graph_name_2: str, result_graph_name: str):
        """
        Merge two graphs using the graph names.

        Args:
            graph_name_1: Name of graph 1
            graph_name_2: Name of graph 2
            result_graph_name: Name of the resulting graph.

        Returns:
            A new graph, the result of merging graph1 and graph2.
        """
        graph_1 = deepcopy(self.graphs[graph_name_1])
        graph_2 = deepcopy(self.graphs[graph_name_2])
        new_nodes, _, nodes_with_changes = self.__get_differences(graph_1, graph_2)
        new_graph_def = graph_1.graph_def
        new_graph_def = self.__add_new_nodes(new_graph_def, new_nodes)
        new_graph_def = self.__copy_updated_nodes(new_graph_def, nodes_with_changes)
        self.add_graph(result_graph_name + '_temp', new_graph_def, graph_1.settings_reader.framework_settings)
        return self.update_graph(graph_name_1, result_graph_name + '_temp')

    def __copy_updated_nodes(self, new_graph_def, nodes_with_changes):
        for node in nodes_with_changes:
            new_graph_def[node.name + '_1'] = node.orig_node_def
        return new_graph_def

    def __add_new_nodes(self, new_graph_def, new_nodes):
        for node in new_nodes:
            new_graph_def[node.name] = node.orig_node_def
        return new_graph_def

    def __get_differences(self, graph_1: TemplateParser, graph_2: TemplateParser):
        graph_1_nodes = self.__get_graph_nodes(graph_1, 'graph1')
        graph_2_nodes = self.__get_graph_nodes(graph_2, 'graph2')
        self.__init_vnodes(graph_1_nodes)
        self.__init_vnodes(graph_2_nodes)
        new_nodes = self.__get_additional_nodes(graph_1_nodes, graph_2_nodes)
        removed_nodes = self.__get_additional_nodes(graph_2_nodes, graph_1_nodes)
        nodes_with_changes = self.__get_nodes_with_changes(graph_1_nodes, graph_2_nodes)
        return new_nodes, removed_nodes, nodes_with_changes

    def __init_vnodes(self, nodes):
        for node in nodes.values():
            node.connect_to_nodes(nodes)

    @staticmethod
    def __get_nodes_with_changes(graph_1_nodes, graph_2_nodes):
        nodes_with_changes = []
        for key, v_node in graph_2_nodes.items():
            if key in graph_1_nodes:
                if v_node.check_differences(graph_1_nodes[key]):
                    nodes_with_changes.append(v_node)
        return nodes_with_changes

    @staticmethod
    def __get_additional_nodes(original_nodes, new_nodes):
        additional_nodes = []
        for node, v_node in new_nodes.items():
            if node not in original_nodes:
                additional_nodes.append(v_node)
        return additional_nodes

    def __get_graph_nodes(self, graph: TemplateParser, graph_name: str):
        graph_def = graph.graph_def
        graph_def = self.__del_keys(graph_def)
        return {key: VirtualNode(key, node, graph_name) for key, node in graph_def.items()}

    @staticmethod
    def __del_keys(graph_def):
        del_keys = ['settings', 'imports', 'outputs']
        for key in del_keys:
            if key in graph_def:
                del graph_def[key]
        return graph_def

    def run(self, **params):
        pass

    def run_all(self):
        """
        Executes all registered graphs.

        Returns:
            None
        """
        for graph_name, graph in self.graphs.items():
            print(f"initializing graph {graph_name}.")
            initialized_graph = graph()
            print(f"executing graph {graph_name}")
            initialized_graph()

    def run_graph(self, graph_name: str):
        """
        Runs the graph with the given name.

        Args:
            graph_name: Name of the graph to run.

        Returns:
            The result given by the graph.
        """
        return self.graphs[graph_name]()

    def init_graph(self, graph: str):
        """
        Initialises the graph.

        Args:
            graph: Name of the graph.

        Returns:
            Initialised graph.
        """
        graph = self.graphs[graph]
        return graph()


class VirtualNode:
    """
    Helper class for graph merging.
    """
    def __init__(self, name, node_definition, graph):
        self.orig_node_def = deepcopy(node_definition)
        self.name = name
        self.graph = graph
        self.setting_nodes = {}
        self.param_nodes = {}
        self.node = deepcopy(node_definition)['node']
        self.node_settings = self.__get_node_settings(deepcopy(node_definition))
        self.params = self.__get_params(deepcopy(node_definition))
        self.update_settings = False
        self.update_params = False
        self.update_setting_nodes = False
        self.update_param_nodes = False
        self.differences_checked = False

    def check_differences(self, new_graph_node):
        if not self.differences_checked:
            self.update_settings = self.__compare(self.node_settings, new_graph_node.node_settings)
            self.update_params = self.__compare(self.params, new_graph_node.params)
            self.update_setting_nodes = self.__compare_nodes(self.setting_nodes, new_graph_node.setting_nodes)
            self.update_param_nodes = self.__compare_nodes(self.param_nodes, new_graph_node.param_nodes)
            self.differences_checked = True
        result = any([self.update_settings, self.update_params, self.update_setting_nodes, self.update_param_nodes])
        return result

    @staticmethod
    def __compare(own_dict, new_graph_dict):
        if not own_dict and not new_graph_dict:
            return False
        elif not own_dict or not new_graph_dict:
            return True
        else:
            for key, value in own_dict.items():
                if key not in new_graph_dict:
                    return True
                elif value != new_graph_dict[key]:
                    return True
            return False

    @staticmethod
    def __compare_nodes(own_nodes, new_graph_nodes):
        if not own_nodes and not new_graph_nodes:
            return False
        elif not own_nodes or not new_graph_nodes:
            return True
        else:
            for key, node in own_nodes.items():
                if key not in new_graph_nodes:
                    return True
                elif node.node != new_graph_nodes[key].node:
                    return True
                elif node.check_differences(new_graph_nodes[key]):
                    return True
            for key, node in new_graph_nodes.items():
                if key not in own_nodes:
                    return True
                elif node.node != own_nodes[key].node:
                    return True
                elif node.check_differences(own_nodes[key]):
                    return True
            return False

    def connect_to_nodes(self, nodes: dict):
        self.setting_nodes = self.__connect_nodes(nodes, self.setting_nodes)
        self.param_nodes = self.__connect_nodes(nodes, self.param_nodes)

    def __connect_nodes(self, nodes: dict, node_paths: dict):
        del_keys = []
        for key, node in node_paths.items():
            for m_node in nodes.values():
                if m_node.name == node:
                    node_paths[key] = m_node
                    break
                elif node in m_node.name and '/' in m_node.name:
                    node_paths[key] = m_node
                    break
            if type(node_paths[key]) is str:
                if node_paths[key] == 'GraphRuntime':
                    del_keys.append(key)
                else:
                    raise ValueError(f'{node} does not exist in {self.graph}.')
        for key in del_keys:
            del node_paths[key]
        return node_paths

    def __get_node_settings(self, node_definition: dict):
        tags = ['//', '/', 'node://', 'node:/']
        if 'node_settings' in node_definition:
            for key, setting in deepcopy(node_definition)['node_settings'].items():
                if type(setting) is str and any([setting.startswith(tag) for tag in tags]):
                    for tag in tags:
                        if setting.startswith(tag):
                            setting = setting.replace(tag, '', 1)
                    self.setting_nodes[key] = setting
                    del node_definition['node_settings'][key]
            return node_definition['node_settings']
        else:
            return None

    def __get_params(self, node_definition: dict):
        tags = ['//', '/', 'node://', 'node:/']
        if 'params' in node_definition:
            for key, param in deepcopy(node_definition)['params'].items():
                if type(param) is str and any([param.startswith(tag) for tag in tags]):
                    for tag in tags:
                        if param.startswith(tag):
                            param = param.replace(tag, '', 1)
                    self.param_nodes[key] = param
                    del node_definition['params'][key]
            return node_definition['params']
        else:
            return None



import numpy as np
from enum import Enum
import cv2

from .math import angle_between_points, angle_relative_to_180
from .standard import tree_search, cal_distance


class BboxFormat(Enum):
    """
    Class used to reference the bounding box bbox_format.
    """
    FOUR_POINT = 0
    TWO_POINT = 1
    XY_WIDTH_HEIGHT = 2


def cut_out_image_area(image, area_bbox):
    """
    Cuts out the specified area from an image and returns it.

    Args:
        image: The original image.
        area_bbox: The area to cut out.

    Returns:
        Cut out area as image.
    """
    max_height, max_width, ordered_bbox = get_area_dimensions(area_bbox)
    # now that we have the dimensions of the new image, construct the set of destination points to obtain a
    # "birds eye view", (i.e. top-down view) of the image, again specifying points in the top-left, top-right,
    # bottom-right, and bottom-left order
    destination_points = np.array([
        [0, 0],
        [max_width - 1, 0],
        [max_width - 1, max_height - 1],
        [0, max_height - 1]], dtype="float32")
    # compute the perspective transform matrix and then apply it
    transform_matrix = cv2.getPerspectiveTransform(ordered_bbox, destination_points)
    transform_matrix = np.array(transform_matrix)
    cut_out_image = cv2.warpPerspective(image, transform_matrix, (max_width, max_height))
    cut_out_image = np.array(cut_out_image)
    # return the cut out image
    return cut_out_image


def get_area_dimensions(area_bbox):
    """
    Gets the area dimension for the given area bbox.

    Args:
        area_bbox: 4 point bbox.

    Returns:
        max_height: Maximum height of the area.
        max_width: Maximum width of the area.
        ordered_bbox: Reordered bounding box as [TL, TR, BR, BL]
    """
    # obtain a consistent order of the points and unpack them individually
    area_bbox = np.array(area_bbox)
    ordered_bbox = order_points(area_bbox)
    (tl, tr, br, bl) = ordered_bbox
    # compute the width of the new image, which will be the maximum distance between bottom-right and bottom-left
    # x-coordiates or the top-right and top-left x-bbox
    width_bottom = np.sqrt(((br[0] - bl[0]) ** 2) + ((br[1] - bl[1]) ** 2))
    width_top = np.sqrt(((tr[0] - tl[0]) ** 2) + ((tr[1] - tl[1]) ** 2))
    max_width = max(int(width_bottom), int(width_top))
    # compute the height of the new image, which will be the maximum distance between the top-right and bottom-right
    # y-bbox or the top-left and bottom-left y-bbox
    height_right = np.sqrt(((tr[0] - br[0]) ** 2) + ((tr[1] - br[1]) ** 2))
    height_left = np.sqrt(((tl[0] - bl[0]) ** 2) + ((tl[1] - bl[1]) ** 2))
    max_height = max(int(height_right), int(height_left))
    return max_height, max_width, ordered_bbox


def four_point_bb_to_2_point(bbox, bbox_format="y1x1y2x2", round_values=False):
    """
    Converts a four point bbox to a two point bbox.

    Args:
        bbox: Bounding box in four point format.
        bbox_format: The format of the bbox as string (for example "y1x1y2x2").

    Returns:
        Two point bbox.
    """
    min_x = np.min([point[0] for point in bbox])
    min_y = np.min([point[1] for point in bbox])
    max_x = np.max([point[0] for point in bbox])
    max_y = np.max([point[1] for point in bbox])

    if round:
        min_x = round(min_x)
        min_y = round(min_y)
        max_x = round(max_x)
        max_y = round(max_y)

    if bbox_format == "y1x1y2x2":
        return [min_y, min_x, max_y, max_x]
    return [min_x, min_y, max_x, max_y]


def two_point_bb_to_4_point(bbox):
    """
    Converts a two point bounding box to a four point bounding box.

    Args:
        bbox: Two point bounding box.

    Returns:
        Four point bbox.
    """
    # convert from (x1,y1,x2,y2) to (x1y1, x2y2)
    if len(bbox) == 4:
        bbox = [[bbox[0], bbox[1]], [bbox[2], bbox[3]]]

    # conversion matches for both formats (x1y1..) also for y1x1y2x2
    x1y1 = bbox[0]
    x4y4 = bbox[1]

    x2y2 = [x4y4[0], x1y1[1]]
    x3y3 = [x1y1[0], x4y4[1]]
    return [x1y1, x2y2, x3y3, x4y4]


def xy_width_height_to_two_point(bbox: list):
    """
    Converts a bounding box with shape [x1,y2,width,height] to a two point bounding box.

    Args:
        bbox: bbox with shape [x1,y2,width,height].

    Returns: bbox with shape [x1y1, x2y2].
    """
    x1y1 = [bbox[0], bbox[1]]
    x2y2 = [bbox[0] + bbox[2], bbox[1] + bbox[3]]
    return [x1y1, x2y2]


def xy_width_height_to_four_point(bbox: list):
    """
    Converts a bounding box with shape [x1,y2,width,height] to a four point bounding box.

    Args:
        bbox: bbox with shape [x1,y2,width,height].

    Returns: four point bbox.
    """
    bbox = xy_width_height_to_two_point(bbox)
    return two_point_bb_to_4_point(bbox)


def bbox_conversion(bbox, bbox_format: BboxFormat, target_bbox_format: BboxFormat) -> np.array:
    """
    This method converts a bounding box to a different bounding box bbox_format.

    Args:
        bbox: the bounding box with the bbox_format of bbox_format
        bbox_format: bbox_format of the bounding box
        target_bbox_format: bbox_format to convert the bounding box to

    Returns: the converted bounding box

    """
    # no conversion necessary
    if bbox_format == target_bbox_format:
        return np.array(bbox)
    # two point to four point
    if bbox_format == BboxFormat.TWO_POINT and target_bbox_format == BboxFormat.FOUR_POINT:
        return np.array(two_point_bb_to_4_point(bbox))
    # point + width, height to four point
    if bbox_format == BboxFormat.XY_WIDTH_HEIGHT and target_bbox_format == BboxFormat.FOUR_POINT:
        return np.array(xy_width_height_to_four_point(bbox))
    # four point to two point
    if bbox_format == BboxFormat.FOUR_POINT and target_bbox_format == BboxFormat.TWO_POINT:
        return np.array(four_point_bb_to_2_point(bbox))


def rotate_bounding_box(bbox, angle):
    """
    Method to rotate the bounding box by a certain angle.

    Args:
        bbox: bounding box in either two point or four point format.
        angle: Angle by which the bounding box should be rotated.

    Returns:
        List containing the 4 points of the bounding box [[x1, y1], [x2, y2], [x3, y3], [x4, y4]].
    """
    bbox, center = get_4_point_bbox_with_center(bbox)
    return get_rotated_bbox(angle, bbox, center)


def get_rotated_bbox(angle, bbox, center):
    """
    Rotates the coordinates by the given angle around the given center.

    Args:
        angle: Angle to rotate by.
        bbox: bounding box.
        center: Center of the rotation as (x, y).

    Returns:
        Rotated bounding box.
    """
    x_center, y_center = center
    rotated_bbox = []
    for point in bbox:
        temp_x = point[0] - x_center
        temp_y = point[1] - y_center
        rotated_x = temp_x * np.cos(angle) - temp_y * np.sin(angle)
        rotated_y = temp_x * np.sin(angle) - temp_y * np.cos(angle)
        final_x = rotated_x + x_center
        final_y = rotated_y + y_center
        rotated_bbox.append([final_x, final_y])
    return rotated_bbox


def get_4_point_bbox_with_center(bbox, center=True):
    """
    Gets the 4 point bbox and its center from the bbox.

    Args:
        bbox: bbox in either two or four point format.
        center: Whether to calculate the center or not (to save performance).

    Returns:
        Four_point_bbox: BBox in four point format.
        Center: Center of the four point bbox if center=True else None.
    """
    x_max, x_min, y_max, y_min = get_x_y_min_max(bbox)
    four_point_bbox = [[x_min, y_min], [x_max, y_min], [x_max, y_max], [x_min, y_max]]
    if center:
        x_center = ((x_max - x_min) / 2) + x_min
        y_center = ((y_max - y_min) / 2) + y_min
        return four_point_bbox, (x_center, y_center)
    else:
        return four_point_bbox, None


def get_x_y_points(bbox):
    """
    Gets the x and y coordinates of the bbox.

    Args:
        bbox: Bounding box in any [(x, y), (x, y), ...] format.

    Returns:
        x_points: X points of the bbox.
        y_points: Y points of the bbox.
    """
    x_points = []
    y_points = []
    for coord in bbox:
        x_points.append(coord[0])
        y_points.append(coord[1])
    return x_points, y_points


def convert_polygon_box_to_4_point_bb(bbox):
    """
    Method to convert a polygon box to a 4 point bbox.

    Args:
        bbox: [[x1, y1], [x2, y2], ..., [xn, yn]]

    Returns: 4-point BBox
    """
    bbox, _ = get_4_point_bbox_with_center(bbox, center=False)
    return bbox

def get_x_y_min_max(bbox):
    """
    Gets the x and y min and max values of the bbox.

    Args:
        bbox: Bounding box in format [[x1, y2], [x2, y2], ..., [xn, yn]]

    Returns:
        x_max: max x value
        x_min: min x value
        y_max: max y value
        y_min: min y value
    """
    x_points, y_points = get_x_y_points(bbox)
    x_max = np.max(x_points)
    x_min = np.min(x_points)
    y_max = np.max(y_points)
    y_min = np.min(y_points)
    return x_max, x_min, y_max, y_min

def flip_bbox(bbox, img_shape, direction):
    """
    Flips a 4 point bounding box in a given direction
    Args:
        bbox: source bounding box
        img_shape: shape of the associated image
        direction: left to right as leftright, top to bottom as topbottom, both as both

    Returns: flipped bounding box

    """
    h, w = img_shape[:2]
    center_x, center_y = int(w/2), int(h/2)
    if direction == "leftright":
        for i in range(len(bbox)):
            bbox[i][0] -= 2 * (bbox[i][0] - center_x)
    elif direction == "topbottom":
        for i in range(len(bbox)):
            bbox[i][1] -= 2 * (bbox[i][1] - center_y)
    elif direction == "both":
        bbox = flip_bbox(bbox, img_shape, "leftright")
        bbox = flip_bbox(bbox, img_shape, "topbottom")
    else:
        raise ValueError("Direction is not supported.")

    return bbox


def to_radian(angle):
    """
    Helper function to convert degree angles into radians.
    Args:
        angle: degree angle

    Returns: radian angle

    """
    return (np.math.pi / 180) * angle


def rotate_bbox_around_center(bbox, cx, cy, angle):
    """
    Rotates a 4 point bounding box around a center point.
    Args:
        bbox: bounding box to rotate
        cx: center x
        cy: center y
        angle: rotation angle in degrees

    Returns: rotated 4 point bounding box as int32

    """
    for i in range(len(bbox)):
        radian = to_radian(angle)
        s = np.math.sin(radian)
        c = np.math.cos(radian)
        px, py = bbox[i]
        xnew = cx + (px - cx) * c - (py - cy) * s
        ynew = cx + (px - cx) * s + (py - cy) * c
        bbox[i] = [xnew, ynew]
    return np.array(bbox).astype(np.int32)


def rescale_bbox(bbox, scale=(1.0, 1.0), offset=(0, 0), skip_none=False):
    """
    Custom function to rescale bounding box.
    Args:
        bbox: source bounding box
        scale: scale factor (xy tuple)
        offset: optional offset for bounding boxes (xy tuple)
        skip_none: skip the function if bbox is none

    Returns: rescaled bounding boxes

    """
    if skip_none and bbox is None:
        return None
    return np.add(np.array(bbox) * scale, offset)


def get_matching_bounding_boxes(bboxes, w, h):
    """
    Removes bounding boxes which are out of range.
    Args:
        bboxes: List of bounding boxes
        w: new image width
        h: new image height

    Returns: prepared list of bboxes

    """
    res = []
    for bbox in bboxes:
        x0 = np.amin([b[0] for b in bbox])
        x1 = np.amax([b[0] for b in bbox])
        y0 = np.amin([b[1] for b in bbox])
        y1 = np.amax([b[1] for b in bbox])

    return np.array([bb for bb in bboxes
                     if np.amin([b[0] for b in bb]) >= 0
                     and np.amax([b[0] for b in bb]) <= w
                     and np.amin([b[1] for b in bb]) >= 0
                     and np.amax([b[1] for b in bb]) <= h])


def convert_triangle_to_quadrangle(box):
    two_point_box = four_point_bb_to_2_point(box)
    return two_point_bb_to_4_point(two_point_box)


def order_points(polygon, clockwise=True, always_return=True):
    if type(polygon) == np.ndarray:
        polygon = polygon.tolist()
    sums = [sum(p) for p in polygon]
    min_v = min(sums)
    min_index = sums.index(min_v)
    set_points = [polygon[min_index]]
    del polygon[min_index]
    if clockwise:
        return tree_search(polygon, set_points, overall_angle, -360, always_return)
    else:
        return tree_search(polygon, set_points, overall_angle, 360, always_return)


def overall_angle(box):
    angles = [angle_relative_to_180(angle_between_points(box[i-1], box[i], box[i+1])) for i in range(len(box[:-1]))]
    angles.append(angle_relative_to_180(angle_between_points(box[-2], box[-1], box[0])))
    return round(sum(angles))


def point_in_box(point, box):
    x_max, x_min, y_max, y_min = get_x_y_min_max(box)
    if point[0] < x_min or point[0] > x_max or point[1] < y_min or point[1] > y_max:
        return False

    box = np.array(order_points(box, always_return=True))
    point = np.array(point)

    # Point calc
    tr1 = 0.5 * cal_distance(box[0], box[3]) * np.linalg.norm(np.cross(box[3] - box[0], box[0] - point)) / \
          np.linalg.norm(box[3] - box[0])
    tr2 = 0.5 * cal_distance(box[3], box[2]) * np.linalg.norm(np.cross(box[2] - box[3], box[3] - point)) / \
          np.linalg.norm(box[2] - box[3])
    tr3 = 0.5 * cal_distance(box[2], box[1]) * np.linalg.norm(np.cross(box[1] - box[2], box[2] - point)) / \
          np.linalg.norm(box[1] - box[2])
    tr4 = 0.5 * cal_distance(box[1], box[0]) * np.linalg.norm(np.cross(box[0] - box[1], box[1] - point)) / \
          np.linalg.norm(box[0] - box[1])

    avg_len = (cal_distance(box[1], box[0]) + cal_distance(box[3], box[2])) / 2
    avg_height = (cal_distance(box[3], box[0]) + cal_distance(box[2], box[1])) / 2
    point = np.array([box[0][0] + avg_len, box[0][1] + avg_height])
    tra1 = 0.5 * cal_distance(box[0], box[3]) * np.linalg.norm(np.cross(box[3] - box[0], box[0] - point)) / \
          np.linalg.norm(box[3] - box[0])
    tra2 = 0.5 * cal_distance(box[3], box[2]) * np.linalg.norm(np.cross(box[2] - box[3], box[3] - point)) / \
          np.linalg.norm(box[2] - box[3])
    tra3 = 0.5 * cal_distance(box[2], box[1]) * np.linalg.norm(np.cross(box[1] - box[2], box[2] - point)) / \
          np.linalg.norm(box[1] - box[2])
    tra4 = 0.5 * cal_distance(box[1], box[0]) * np.linalg.norm(np.cross(box[0] - box[1], box[1] - point)) / \
          np.linalg.norm(box[0] - box[1])

    quadrangle_area = sum([tra1, tra2, tra3, tra4])
    calc_p_area = sum([tr1, tr2, tr3, tr4])

    return quadrangle_area >= calc_p_area


def move_in_any_box(point, boxes):
    point = [point[0], point[1]]
    distances = [get_4_point_bbox_with_center(box)[1] for box in boxes]
    box = boxes[distances.index(min(distances))]
    x_max, x_min, y_max, y_min = get_x_y_min_max(box)
    if point[0] < x_min or point[0] > x_max:
        point[0] = x_min if abs(x_min - point[0]) < abs(x_max - point[0]) else x_max
    if point[1] < y_min or point[1] > y_max:
        point[1] = y_min if abs(y_min - point[1]) < abs(y_max - point[1]) else y_max
    return point


import psutil
import subprocess

def check_if_process_running(processName):
    '''
    Check if there is any running process that contains the given name processName.
    '''
    #Iterate over the all the running process
    for proc in psutil.process_iter():
        try:
            # Check if process name contains the given name string.
            if processName.lower() in proc.name().lower():
                return True
        except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
            pass
    return False


def start_process_if_not_running(process_name, *args):
    if not check_if_process_running(process_name):
        subprocess.Popen([process_name, *args])



import json
import os
from concurrent.futures.thread import ThreadPoolExecutor
import threading
from typing import Union

from grandpa.node import Node
from grandpa.utils.standard import do_import, filter_kwargs


class SettingNode(Node):
    """
    This class can be used to create a node that validates a user setting. Requires a SettingReaderNode as
    param 'settings'.
    """
    def __init__(self, setting, default=None, setting_type=None, val_func=None, **kwargs):
        self.setting = setting
        self.default = default
        self.setting_type = setting_type
        self.val_func = val_func
        super().__init__(**kwargs)

    def __check_setting_value(self, setting_value):
        """

        Args:
            setting_value: Value of the setting.

        Returns:
            Settings value if it passes all checks, else the default value given at init.
        """
        if not self.__check_type(setting_value):
            return self.default
        elif not self.__get_val_func_affirmation(setting_value):
            return self.default
        else:
            return setting_value

    def __check_type(self, setting_value):
        """

        Args:
            setting_value: Value of the setting.

        Returns:
            True if the setting matches the expected type, else False.
        """
        if self.setting_type is None:
            return True
        elif type(setting_value) == self.setting_type:
            return True
        else:
            return False

    def __get_val_func_affirmation(self, setting_value):
        """

        Args:
            setting_value: Value of the setting.

        Returns:
            True if the setting passes the val func (or there is no val func), else False.
        """
        if self.val_func is not None:
            return self.val_func(setting_value)
        else:
            return True

    def run(self, **params):
        """
        Executes the node. Shouldn't be called directly. Instead, use __call__.

        Args:
            **params: Not implemented.

        Returns:
            Settings value if it passes all checks, else the default value given at init.
        """
        setting_value = self.setting
        return self.__check_setting_value(setting_value)


class SettingReaderNode(Node):
    """
    This class will read the Settings.
    """
    def __init__(self, setting_path: Union[dict, str], **kwargs):
        self.__init_settings(setting_path)
        super().__init__(**kwargs)
        self.add_param('setting')

    def __init_settings(self, setting_path: Union[dict, str]):
        """
        Initialises the settings.

        Args:
            setting_path: Path to the settings file or a dict containing the settings.

        Returns:
            None
        """
        if type(setting_path) is str:
            self.__read_setting_from_file(setting_path)
        elif type(setting_path) is dict:
            self.framework_settings = setting_path
        else:
            raise NotImplementedError(f'Type {type(setting_path)} is not implemented, it must be either dict or '
                                      f'string.')

    def __read_setting_from_file(self, setting_path: str):
        """
        Reads the settings from a file.

        Args:
            setting_path: Path to the settings file.

        Returns:
            None
        """
        assert os.path.exists(setting_path), f'File {setting_path} does not exist.'
        with open(setting_path, 'r') as file:
            self.framework_settings = json.load(file)

    def __parse_dict_reference(self, setting: str):
        """

        Args:
            setting: Framework internal settings path. E.g. 'network/backbone/input_shape'

        Returns:
            The value of the setting at the given path.
        """
        target_setting = self.framework_settings
        if setting == '' or setting == "*":
            return target_setting
        for part in setting.split('/'):
            target_setting = target_setting[part]
        return target_setting

    def run(self, **params):
        """
        Executes the node. Shouldn't be called directly. Instead, use __call__.

        Args:
            **params: Must include a parameter setting that specifies the framework internal settings path, e.g.
            'network/backbone/input_shape'.

        Returns:
            Settings value for the given settings path.
        """
        assert hasattr(self, 'settings'), 'Something must have gone horribly wrong, since your SettingReaderNode is ' \
                                          'lacking the "settings" attribute that should always be created at init. ' \
                                          'As this is most likely a source code problem, please check the code and ' \
                                          'contact the developers.'
        return self.__parse_dict_reference(params['setting'])


class QueueWrapper(Node):
    """
    Wraps a node as queue. Will constantly generate data until a upper limit (based on pc specs) is reached using the
    framework multiprocessing. Calling this node will return an entry from the queue.
    """
    def __init__(self, target_node, **kwargs):
        super().__init__(**kwargs)
        self.worker_queue = self.switch.add_worker_queue(target=target_node)

    def run(self, **params):
        return self.worker_queue.get()


class FuncExecNode(Node):
    """
    This node can be understood as a function wrapper. The function must be passed as func at init, all parameters
    required for the function need to be passed at call, either directly (call with params) or via nodes
    (self.add_node(name, node)).
    """
    def __init__(self, func, **kwargs):
        if not callable(func):
            func = do_import(func)
        self.func = func
        super().__init__(**kwargs)

    def run(self, **params):
        filtered_kwargs = filter_kwargs(self.func, params)
        return self.func(**filtered_kwargs)


class FuncCacheNode(Node):
    """
    This Node wraps a function the same way as FuncExecNode, except it only calls the function once and caches the
    result.
    """
    def __init__(self, func, **kwargs):
        self.func = func
        self.value = None
        self.lock = threading.Lock()
        super().__init__(**kwargs)

    def run(self, **params):
        self.lock.acquire()
        if not self.value:
            self.value = self.func(**params)
        self.lock.release()
        return self.value


class ThreadedMultiNode(Node):
    """
    This node can be used to call another node multiple times with different parameters in parallel (threading).
    Deprecated.
    """
    def __init__(self, node, **kwargs):
        self.node = node
        super().__init__(**kwargs)
        self.add_param('settings')

    def run(self, settings):
        with ThreadPoolExecutor() as tpe:
            threads = {}
            for name, setting_set in settings.items():
                threads[name] = tpe.submit(self.node, **setting_set)
            return {name: t.result() for name, t in threads.items()}


class ListWrapperNode(Node):
    """
    This node can be used to wrap a list as a node.
    """
    def __init__(self, sub_nodes: list, **kwargs):
        self.sub_nodes = sub_nodes
        super().__init__(**kwargs)

    def run(self):
        return [node() for node in self.sub_nodes]


class DictWrapperNode(Node):
    """
    This node can be used to wrap a dict as a node.
    """

    def __init__(self, sub_nodes: dict, **kwargs):
        self.sub_nodes = sub_nodes
        super().__init__(**kwargs)

    def run(self):
        return {key: node() for key, node in self.sub_nodes.items()}


class ValueWrapperNode(Node):
    """
    This node can be used in case you want to store a fixed value in a node. It will return the stored value every
    time it is called.
    """
    def __init__(self, value, **kwargs):
        self.value = value
        super().__init__(**kwargs)

    def run(self):
        return self.value


class ResultWrapperNode(Node):
    """
    This node is used in order to wrap the result of multiple nodes into one dict. All nodes must be added via
    add_node(name, node) to the ResultWrapperNode.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def run(self, **params):
        return {name: value for name, value in params.items()}



import multiprocessing
import os
import threading
import time
from queue import Queue
from threading import Thread

from psutil import virtual_memory

from grandpa.utils.standard import print_warning, check_debug


class MultiprocessingManager:
    """
    Manages the multiprocessing of the framework.
    """
    def __init__(self):
        self.__task_queue = Queue()
        self.__worker_queue_list = []
        self.__init_workers()

    def add_worker_queue(self, worker_queue):
        """
        Registers a worker Queue.

        Args:
            worker_queue: Worker Queue Object to register.

        Returns:
            None
        """
        self.__worker_queue_list.append(worker_queue)

    def add_task(self, task):
        """
        Adds a task to be processed.

        Args:
            task: Task Object.

        Returns:
            None
        """
        self.__task_queue.put(task)

    def __init_workers(self):
        debug = os.environ["project_x_debug"] if "project_x_debug" in os.environ else False
        if debug == "true":
            worker_count = 1
        else:
            worker_count = multiprocessing.cpu_count() - 2
        for _ in range(worker_count):
            t = Thread(target=self.__work)
            t.daemon = True
            t.start()

    def __work(self):
        while True:
            try:
                if self.__task_queue.qsize() > 0:
                    task = self.__task_queue.get()
                    if task.lock.locked():
                        continue
                    else:
                        task.get_result()
                elif len(self.__worker_queue_list) > 0:
                    task_queue = self.__get_task_queue()
                    if task_queue is None:
                        time.sleep(0.1)
                    else:
                        task_queue.generate()
                else:
                    time.sleep(0.1)
            except Exception as e:
                if check_debug():
                    raise e
                else:
                    print_warning(str(e))

    def __get_task_queue(self):
        target_queue = None
        target_queue_size = float('inf')
        for t_queue in self.__worker_queue_list:
            if t_queue.qsize() / t_queue.target_size < target_queue_size and t_queue.qsize() < t_queue.target_size:
                target_queue_size = t_queue.qsize() / t_queue.target_size
                target_queue = t_queue
        return target_queue


class Task:
    """
    Task Class for Multiprocessing. Will execute once.
    """
    def __init__(self, multiprocessing_manager: MultiprocessingManager, target, *args, **kwargs):
        self.target = target
        self.args = args
        self.kwargs = kwargs
        self.lock = threading.Lock()
        self.result = NoData()
        multiprocessing_manager.add_task(self)

    def get_result(self):
        self.lock.acquire()
        if type(self.result) == NoData:
            self.result = self.target(*self.args, **self.kwargs)
        self.lock.release()
        return self.result


class WorkerQueue(Queue):
    """
    Class for registering a worker queue. Runs constantly until target_size is reached.
    """
    def __init__(self, multiprocessing_manager: MultiprocessingManager, target,
                 target_size: int = int(virtual_memory().total / 1000000000), split_results: bool = False,
                 batches_per_run: int = 1, *args, **kwargs):
        super().__init__()
        self.target = target
        self.args = args
        self.kwargs = kwargs
        self.target_size = target_size
        self.split_results = split_results
        self.batches_per_run = batches_per_run
        self.__pending_batches = 0
        self.static_workers = []
        self.__init_static_workers(1)
        multiprocessing_manager.add_worker_queue(self)

    def __init_static_workers(self, worker_num: int):
        debug = os.getenv("project_x_debug", False)
        if not debug == "true":
            for _ in range(worker_num):
                t = Thread(target=self.__worker_thread)
                t.daemon = True
                t.start()
                self.static_workers.append(t)

    def __worker_thread(self):
        if self.qsize() < self.target_size:
            self.generate()
        else:
            time.sleep(1)

    def get(self, block=True, timeout=None):
        debug = os.getenv("project_x_debug", False)
        if self.qsize() > 0:
            return super().get(block, timeout)
        else:
            try:
                self.generate()
            except Exception as e:
                if debug:
                    raise e
                else:
                    print(e)
            return self.get(block, timeout)

    def generate(self):
        self.__pending_batches += self.batches_per_run
        try:
            result = self.target(*self.args, **self.kwargs)
            if self.split_results:
                for r in result:
                    self.put(r)
            else:
                self.put(result)
        except Exception as e:
            if check_debug():
                raise e
            else:
                print_warning(str(e))
                return self.generate()
        self.__pending_batches -= self.batches_per_run

    def qsize_old(self) -> int:
        return super().qsize() + self.__pending_batches


class NoData:
    """
    Helper class. Symbolizes that a Task has not yet been completed.
    """
    pass



from typing import Union
import re

from psutil import virtual_memory

from grandpa.multiprocessing_manager import MultiprocessingManager, Task, WorkerQueue


class Switch:
    """
    Used to connect nodes to each other.
    """
    def __init__(self, address, node, main_router=None):
        self.address = address
        self.factory = node
        self.default_gateway = main_router.get_parent(address) if address and main_router else None
        self.address_table = {}
        self.main_router = main_router
        if address:
            if self.default_gateway:
                self.default_gateway.register(self.factory, self.address)
            else:
                if main_router:
                    main_router.register(self.factory, self.address)

    def execute_task(self, target, *args, **kwargs) -> Task:
        """
        Adds a task to execute by the MultiProcessingManager.

        Args:
            target: Callable target.
            *args: Call args.
            **kwargs: Call kwargs.

        Returns:
            Task Object, that can be used to get the result.
        """
        return self.main_router.execute_task(target, *args, **kwargs)

    def add_worker_queue(self, target, target_size: int = int(virtual_memory().total / 1000000000),
                         split_results: bool = False, *args, **kwargs) -> WorkerQueue:
        """
        Adds a worker queue to the MultiProcessingManager.

        Args:
            target: Callable target of the queue.
            target_size: Maximum size of the queue.
            split_results: If True, results will be split (e.g., ['A', 'B', 'C'] would add 3 entries to the queue.
            *args: Args to use when calling target.
            **kwargs: Kwargs to use when calling target.

        Returns:
            Worker Queue object.
        """
        return self.main_router.add_worker_queue(target, target_size=target_size, split_results=split_results, *args,
                                                 **kwargs)

    def register(self, target, name=None):
        if name:
            self.address_table[f'{self.address}/{name}'] = target
        else:
            self.address_table[f'{self.address}/{target.name}'] = target

    def delete_object_at_path(self, path):
        target = self.get_instruction(path)
        del target
        if path in self.address_table:
            del self.address_table[path]

    def register_foreign(self, path, target):
        self.address_table[path] = target

    def route(self, path, return_value, call_trace: list = None):
        """
        Public call for the route function.

        Args:
            path: Path to route to.
            return_value: Whether to return the value of the object (True) or the object stored at this location.
            call_trace: The switch / router this function is called at will share its address table with the
            routers / switches in this list.

        Returns:
            The object or the object value stored at the location given by the path.
        """
        if call_trace:
            for switch in call_trace:
                self.__share_address_table(switch)
        return self.__route(path, return_value)

    def __share_address_table(self, switch):
        for address in self.address_table:
            if address not in switch.address_table:
                switch.register_foreign(address, self.__route(address, False))

    def __route(self, path: str, return_value):
        """
        Routes the request. Returns the requested value if the switch can directly access it,
        or forwards the request to the next switch in the direction of the target.

        Args:
            path: Path of the requested value.
            return_value: Whether to return the value of the object (True) or the object stored at this location.

        Returns:
            The object or the object value stored at the location given by the path.
        """
        if path.startswith('//'):
            path = path[1:]
        if path in self.address_table:
            return self.__get_routing_target(path, return_value)
        elif self.__get_closest_switch(path):
            return self._forward_to_switch(path, return_value, self.__get_closest_switch(path))
        else:
            if self.default_gateway:
                return self.default_gateway.route(path, return_value)
            else:
                raise RoutingError(route=path)

    def _forward_to_switch(self, path, return_value, switch):
        if type(switch) is Switch:
            return switch.route(path, return_value)
        elif type(switch) is dict:
            return self.__resolve_dict_path(path)

    def __resolve_dict_path(self, path):
        path = path.replace(self.address, '')[1:]
        resolve_dict = self.get_value(self.address + '/' + path.split('/')[0])
        in_dict_path = path.replace(path.split('/')[0], '')[1:]
        for key in in_dict_path.split('/'):
            resolve_dict = resolve_dict[key]
        return resolve_dict

    def __get_closest_switch(self, path: str):
        matching_switches = [self.address_table[s_path] for s_path in self.address_table
                             if re.match(s_path + '/.*', path)]
        return max(matching_switches) if matching_switches else None

    def __get_routing_target(self, path, return_value):
        """
        Returns the value or object at the path location.

        Args:
            path: Path where the object is stored.
            return_value: Whether to return the object or the object value.

        Returns:
            The object or the object value stored at the location given by the path.
        """
        if return_value and callable(self.address_table[path]):
            return self.address_table[path]()
        else:
            return self.address_table[path]

    def get_instruction(self, path):
        return self.__route(path, return_value=False)

    def get_value(self, path):
        return self.__route(path, return_value=True)

    def get_all_targets(self):
        return list(self.address_table.values())

    def get_filtered_targets(self, target_type_filter: Union[type, list] = None) -> list:
        if not target_type_filter:
            return self.get_all_targets()

        if type(target_type_filter) == type:
            target_type_filter = [target_type_filter]
            return list(p for p in self.address_table.values() if type(p) in target_type_filter)

    def get_target_dependencies(self, target_type_filter: Union[type, list] = None):
        targets = self.get_filtered_targets(target_type_filter)
        return [p for p in targets if "/" in p.value]


class Router(Switch):
    def __init__(self, default_gateway=None, routing_table=None):
        super().__init__('', None, None)
        self.default_gateway = default_gateway
        self.routing_table = routing_table
        self.multiprocessing_manager = MultiprocessingManager()

    def execute_task(self, target, *args, **kwargs) -> Task:
        return Task(self.multiprocessing_manager, target, *args, **kwargs)

    def add_worker_queue(self, target, target_size: int = int(virtual_memory().total / 1000000000),
                         split_results: bool = False, *args, **kwargs) -> WorkerQueue:
        return WorkerQueue(self.multiprocessing_manager, target, target_size=target_size, split_results=split_results,
                           *args, **kwargs)

    def get_parent(self, address: str):
        if '/' not in address:
            return self
        else:
            address = address.rsplit('/', 1)[0] if not address.endswith('/') else address.rsplit('/', 2)[0]
            try:
                return self.get_instruction(address).switch if address != '' else self
            except RoutingError:
                return self.get_parent(address)

    @staticmethod
    def __convert_to_relative_path(path):
        assert path[0] == '/', f'Error: {path} is not a global path. Did you want to search for /{path}?'
        path = path.replace('/', '', 1)
        return path


class RoutingError(Exception):
    def __init__(self, route):
        self.message = f'Route {route} does not exist - it might not have been created.'
        super().__init__(self.message)



from distutils.core import setup


install_requires=[
   "setuptools",
    "wheel",
    "imageio",
    "numpy",
    "opencv-python",
    "Pillow",
    "psutil",
    "tqdm"
]


setup(
    name='grandpa',
    version='0.4.3',
    packages=['grandpa', 'grandpa.utils'],
    url='https://dev.azure.com/matthiasrieger/ImageRecognition/_git/ir-graml-core',
    license='MIT License',
    author='Bizerba AI Team',
    author_email='pascal.iwohn@bizerba.com',
    description='',
    install_requires=install_requires,
    package_dir={"": "src"}
)



import os
import random
from typing import Union

from grandpa.built_in_nodes import ValueWrapperNode, ResultWrapperNode, SettingReaderNode, \
    ListWrapperNode, DictWrapperNode, FuncExecNode
from grandpa.utils.standard import do_import, print_warning
from grandpa.routing import Router
import json


class TemplateParser:
    def __init__(self, template: dict, settings, router: Router):
        self.main_router = router
        self.settings_reader = self.__init_settings_reader(settings)
        self.template = template
        self.outputs = self.__init_outputs(template)
        self.graph_def = self.__do_imports(template)
        self.initialized_nodes = []

    def __call__(self):
        return self.parse()

    def parse(self):
        for node_address, node_definition in self.graph_def.items():
            if node_address not in self.initialized_nodes:
                self.init_node(node_address, node_definition)
        if self.outputs:
            return self.__get_output_node()
        else:
            return self.__auto_get_output_node()

    def __do_imports(self, graph_def):
        if "imports" in graph_def:
            assert type(graph_def["imports"]) is list, f"Imports must be given to the parser as a list of graph " \
                                                       f"definitions, not a single graph definition. If you want to " \
                                                       f"pass a single graph definition, use [graph_def]."
            for imp in graph_def["imports"]:
                graph_def_imp = self.__do_imports(imp)
                for node, definition in graph_def_imp.items():
                    if node not in graph_def:
                        graph_def[node] = definition
        graph_def = self.__del_obsolete_params(graph_def)
        return graph_def

    @staticmethod
    def __del_obsolete_params(graph):
        keys = ["settings", "outputs", "imports"]
        for key in keys:
            if key in graph:
                del graph[key]
        return graph

    def __init_outputs(self, graph_def):
        if 'outputs' in graph_def:
            outputs = graph_def['outputs']
            return outputs
        else:
            return None

    def __get_output_node(self):
        if type(self.outputs) is str:
            return self.__get_param_node(self.outputs, '')
        elif type(self.outputs) is dict:
            return self.__create_result_wrapper(self.outputs)
        else:
            raise TypeError(f'{type(self.outputs)} is not allowed as type to declare outputs. '
                            f'Use either string for a single output or dict for multiple outputs.')

    def __auto_get_output_node(self):
        output_node = None
        for node in self.graph_def:
            dependency_count = self.__calc_dependency_count(node)
            if dependency_count == 0:
                output_node = node
        if not output_node.startswith('//'):
            output_node = '//' + output_node
        return self.__get_param_node(output_node, '')

    def __calc_dependency_count(self, node):
        dependency_count = 0
        for node_name, graph_params in self.graph_def.items():
            for key in ['node_settings', 'params']:
                if key in graph_params:
                    for key_node in graph_params[key].values():
                        if type(key_node) is str and self.__get_node_name(key_node, node_name) == node:
                            dependency_count += 1
        return dependency_count

    def __create_result_wrapper(self, output_def: dict):
        result_wrapper = ResultWrapperNode(address=f'GenResultWrapper{random.getrandbits(128)}',
                                           main_router=self.main_router)
        for name, node in output_def.items():
            result_wrapper.add_node(name, self.__get_param_node(node, ''))
        return result_wrapper

    def init_node(self, node_address, node_definition, node_call_stack=None):
        if node_call_stack is None:
            node_call_stack = [node_address]
        else:
            self.__check_for_circular_definition(node_address, node_call_stack)
            node_call_stack.append(node_address)
        self.__make_node_ready_for_init(node_call_stack, node_definition, node_address)
        try:
            node = self.__import_node(node_definition['node'])
        except ModuleNotFoundError as e:
            raise GraphDefinitionError(
                f"Node {node_address} could not be imported from path {node_definition['node']}. "
                f"Please check the import path. Error raised: {e}")
        if 'node_settings' in node_definition:
            node_settings = self.__get_node_settings(node_address, node_definition['node_settings'])
        else:
            node_settings = {}
        initialized_node = node(**node_settings, address=node_address, main_router=self.main_router)
        if 'params' in node_definition:
            self.add_nodes(initialized_node, node_definition['params'], node_address)
        self.initialized_nodes.append(node_address)

    def add_nodes(self, initialized_node, params, node_address):
        for name, param in params.items():
            param_node = self.__get_param_node(param, node_address)
            initialized_node.add_node(name, param_node)
        return initialized_node

    def __get_param_node(self, param, node_address: str):
        if type(param) is str:
            node_get_dict = {
                '//': (self.main_router.get_instruction, {"path": param}),
                '/': (self.main_router.get_instruction, {"path": '//' + node_address + param}),
            }
            node_create_dict = {
                'settings:': (self.settings_reader, {"setting": param.replace('settings:', '')}),
                'node:': (self.main_router.get_instruction, {"path": param.replace('node:', '')
                if param.replace('node:', '').startswith('//') else '//' + node_address + param.replace('node:', '')}),
                'func:': (FuncExecNode,
                          {"func": param.replace('func:', ''), "address": f"GenFuncExecNode/{random.getrandbits(128)}",
                           "main_router": self.main_router})
            }
            for tag, node in node_get_dict.items():
                if param.startswith(tag):
                    return node[0](**node[1])
            for tag, node in node_create_dict.items():
                if param.startswith(tag):
                    return ValueWrapperNode(node[0](**node[1]), address=f"GenValueWrapper/{random.getrandbits(128)}",
                                            main_router=self.main_router)
        elif type(param) is list:
            return ListWrapperNode([self.__get_param_node(p, node_address) for p in param],
                                   address=f"GenListWrapper/{random.getrandbits(128)}", main_router=self.main_router)
        elif type(param) is dict:
            return DictWrapperNode({key: self.__get_param_node(p, node_address) for key, p in param.items()},
                                   address=f"GenDictWrapper/{random.getrandbits(128)}", main_router=self.main_router)
        return ValueWrapperNode(param, address=f"GenValueWrapper/{random.getrandbits(128)}",
                                main_router=self.main_router)

    def __get_node_settings(self, node_address, def_node_settings: dict):
        node_settings = {}
        for name, setting in def_node_settings.items():
            if type(setting) is str and setting.startswith('**'):
                node_settings = {**self.__resolve_setting(node_address, setting), **node_settings}
            else:
                node_settings[name] = self.__resolve_setting(node_address, setting)
        return node_settings

    def __resolve_setting(self, node_address: str, setting):
        if type(setting) is str:
            return self.__resolve_settings_string(node_address, setting)
        elif type(setting) is list:
            return [self.__resolve_setting(node_address, sub_setting) for sub_setting in setting]
        elif type(setting) is dict:
            return {key: self.__resolve_setting(node_address, value) for key, value in setting.items()}
        else:
            return setting

    def __resolve_settings_string(self, node_address, setting):
        if setting.startswith('**'):
            setting.replace('**', '')
            return_dict = True
        else:
            return_dict = False
        tag_action_dict = {
            '//': (self.main_router.get_value, {"path": setting}),
            '/': (self.main_router.get_value, {"path": '//' + node_address + setting}),
            'settings:': (self.settings_reader, {"setting": setting.replace('settings:', '')}),
            'node:': (self.main_router.get_instruction, {"path": setting.replace('node:', '')
            if setting.replace('node:', '').startswith(
                '//') else '//' + node_address + setting.replace('node:', '')}),
            'func:': (FuncExecNode, {"func": setting.replace('func:', ''),
                                     "address": f"GenFuncExecNode/{random.getrandbits(128)}",
                                     "main_router": self.main_router})
        }
        for tag, action in tag_action_dict.items():
            if setting.startswith(tag):
                return action[0](**action[1])
        if return_dict:
            return {**setting}
        else:
            return setting

    @staticmethod
    def __import_node(node: str):
        if '.' in node:
            return do_import(node)
        else:
            return do_import('grandpa.built_in_nodes.' + node)

    @staticmethod
    def __check_for_circular_definition(node_address, node_call_stack):
        if node_address in node_call_stack:
            # raise CircularGraphError(node_call_stack)
            pass

    def __make_node_ready_for_init(self, node_call_stack, node_definition: dict, node_address):
        for node_def_key in ['node_settings', 'params']:
            if node_def_key in node_definition:
                self.__ensure_that_entry_is_dict(node_definition[node_def_key], node_def_key)
                required_uninitialised_nodes = self.__get_required_uninitialised_notes(node_definition[node_def_key],
                                                                                       node_address)
                if required_uninitialised_nodes:
                    for req_uninit_node in required_uninitialised_nodes:
                        self.init_node(req_uninit_node, self.graph_def[req_uninit_node], node_call_stack)

    def __get_required_uninitialised_notes(self, requirements: dict, node_address):
        required_uninitialised_nodes = []
        for req in requirements.values():
            self.__append_node_name_if_required(node_address, req, required_uninitialised_nodes)
        return required_uninitialised_nodes

    def __append_node_name_if_required(self, node_address, requirement, required_uninitialised_nodes):
        if type(requirement) is str:
            node_name = self.__get_node_name(requirement, node_address)
            if self.__is_node(requirement, node_name):
                if node_name not in self.initialized_nodes:
                    required_uninitialised_nodes.append(node_name)
        elif type(requirement) is list:
            for req in requirement:
                self.__append_node_name_if_required(node_address, req, required_uninitialised_nodes)
        elif type(requirement) is dict:
            for k, req in requirement.items():
                self.__append_node_name_if_required(node_address, req, required_uninitialised_nodes)

    def __is_node(self, req, node_name):
        if 'node:' in req:
            self.__ensure_that_node_is_in_graph_def(node_name, req)
            return True
        elif '/' in req:
            return self.__check_if_name_in_graph_def(node_name)
        else:
            return False

    def __check_if_name_in_graph_def(self, node_name):
        if node_name in self.graph_def:
            return True
        else:
            return False

    def __ensure_that_node_is_in_graph_def(self, node_name, req):
        if node_name == "runtime":
            return

        assert node_name in self.graph_def, \
            f'Node {node_name} was directly referenced in graph def as {req}, but there is no node with the ' \
            f'name {node_name} defined.'

    @staticmethod
    def __get_node_name(node_ref: str, node_address: str):
        if type(node_ref) is str:
            node_ref = node_ref.replace('node:', '')
            if not node_ref.startswith('//'):
                return node_address + node_ref
            else:
                return node_ref[:2].replace('/', '') + node_ref[2:]
        else:
            return None

    @staticmethod
    def __ensure_that_entry_is_dict(entry, name: str):
        assert type(entry) is dict, f'{name} in graph definition must be defined as a dict, not a {type(entry)}.'

    def __init_settings_reader(self, settings: Union[dict, str]):
        if type(settings) is str:
            if not os.path.exists(settings):
                print_warning(f'Graph parser tried to load settings form file {settings}, '
                              f'but the file does not exist. Please check the file path.')
                return None
            return SettingReaderNode(json.load(open(settings)), address='SettingsReader', main_router=self.main_router)
        elif type(settings) is dict:
            return SettingReaderNode(settings, address='SettingsReader', main_router=self.main_router)
        elif settings is not dict:
            print_warning(f'Graph parser was called with settings of type {type(settings)} - however, '
                          f'this type is not supported, the settings must be either given as a file path or a dict.')
        else:
            print_warning(f'No settings file was handed to the graph parser - all settings will therefore be '
                          f'initialized with their default values.')

        return None


class CircularGraphError(Exception):
    def __init__(self, call_stack: list):
        self.message = f"You have successfully created a circular graph. This means the graph you defined is " \
                       f"essentially like a dog, it will attempt to chase its tail but never catch up to it. The " \
                       f"call order the parser generated from your graph definition was {' -> '.join(call_stack)}, " \
                       f"so in order to fix this problem, you will need to change the node_settings or params of one " \
                       f"of these nodes."
        super().__init__(self.message)


class CircularImportError(Exception):
    def __init__(self, call_stack: list):
        self.message = f"You have successfully created a circular import. This means the imports you defined are " \
                       f"importing each other, creating an endless loop. The import order careated was " \
                       f"{' -> '.join(call_stack)}, so in order to fix this problem, you will need to remove one of " \
                       f"these imports."
        super().__init__(self.message)


class GraphDefinitionError(Exception):
    def __init__(self, message):
        super().__init__(message)



import multiprocessing
import threading

from grandpa.routing import Switch
from grandpa.utils.standard import print_warning, filter_kwargs
import random


class NodeCache:
    """
    Class used for result caching.
    """
    def __init__(self, cache_size=multiprocessing.cpu_count()):
        self.cached_values = []
        self.cache_ids = []
        self.cache_size = cache_size

    def save(self, id, value):
        """
        Saves data to the cache.

        Args:
            id: ID of the data point.
            value: Value of the data point.

        Returns:
            None
        """
        if len(self.cache_ids) >= self.cache_size:
            del self.cached_values[0]
            del self.cache_ids[0]
        self.cached_values.append(value)
        self.cache_ids.append(id)

    def contains(self, id):
        """
        Checks if the given ID exists in the cache.

        Args:
            id: ID to search for.

        Returns:
            True if the ID is cached, else False.
        """
        return id in self.cache_ids

    def __call__(self, id):
        """
        Reads the ID's value from the cache.

        Args:
            id: ID to read.

        Returns:
            Value of the ID or None, if ID doesn't exist.
        """
        if id in self.cache_ids:
            return self.cached_values[self.cache_ids.index(id)]


class Node:
    """
    Parent class for all graph Nodes. All Nodes written by the user must inherit this class.
    """
    def __init__(self, address, main_router):
        self.nodes = {}
        self.address = address
        self.params = {}
        self.settings = {}
        self.switch = Switch(address, self, main_router)
        self.thread_lock = threading.Lock()
        print(f'Created new node {self.address}')
        self.cache = NodeCache()

    def add_node(self, name, node):
        """
        Adds a dependency node to this node.

        Args:
            name: Name as which the value should be passed in at run.
            node: The dependency node.

        Returns:
            None
        """
        self.__check_if_node_already_exists(name)
        self.nodes[name] = '//' + node.address

    def add_param(self, name, default='__required__'):
        self.params[name] = default

    def add_setting(self, name, value):
        self.settings[name] = value
        return value

    def _validate_params(self, params):
        for req_param, default in self.params.items():
            if req_param not in params:
                if default != '__required__':
                    params[req_param] = default
                else:
                    raise ValueError(f'The required param {req_param} for {self.address} was not set in graph '
                                     f'definition and no default was provided.')
        return params

    def has_param(self, name):
        return name in self.params

    def __check_if_node_already_exists(self, name):
        if name in self.nodes:
            print_warning(f'Parameter {name} for node {self.address} has already been set - '
                          f'previous value will be overridden.')

    def __call_sub_nodes_threaded(self, call_params, nodes, call_id):
        tasks = {}
        for name, node in nodes.items():
            tasks[name] = self.switch.execute_task(node, **{"call_id": call_id, **call_params})
        result = {name: t.get_result() for name, t in tasks.items()}
        return result

    def run(self, *args, **kwargs):
        raise NotImplementedError('run() must be implemented in sub classes of Node.')

    def run_with_call_id(self, *args, call_id=None, **call_params):
        run_nodes = {name: self.switch.get_instruction(node) for name, node in self.nodes.items() if
                     name not in call_params}
        params = self.__call_sub_nodes_threaded(call_params=call_params, nodes=run_nodes, call_id=call_id)
        filtered_call_params = filter_kwargs(self.run, call_params)
        params = {**params, **filtered_call_params}
        # params = self._validate_params(params)
        return self.run(*args, **params)

    def __call__(self, *args, call_id=None, **call_params):
        """
        Executes the node.

        Args:
            *args: Arguments to pass to this node.
            call_id: ID of the call. Usually assigned automatically by the framework.
            **call_params: Parameters to pass to the node at call.

        Returns:
            Result delivered by the node.
        """
        self.thread_lock.acquire()
        if call_id is None:
            call_id = random.getrandbits(128)
        if self.cache.contains(call_id):
            ret_value = self.cache(call_id)
        else:
            ret_value = self.run_with_call_id(*args, call_id=call_id, **call_params)
            self.cache.save(call_id, ret_value)
        self.thread_lock.release()
        return ret_value
